from __future__ import annotations

import pytest

from daft.expressions import col
from daft.recordbatch import MicroPartition


def test_binary_length() -> None:
    table = MicroPartition.from_pydict(
        {
            "col": [
                b"foo",  # Basic ASCII
                None,  # Null value
                b"",  # Empty string
                b"Hello\xe2\x98\x83World",  # UTF-8 character in middle
                b"\xf0\x9f\x98\x89test",  # UTF-8 bytes at start
                b"test\xf0\x9f\x8c\x88",  # UTF-8 bytes at end
                b"\xe2\x98\x83\xf0\x9f\x98\x89\xf0\x9f\x8c\x88",  # Multiple UTF-8 sequences
                b"Hello\x00World",  # Null character
                b"\xff\xfe\xfd",  # High bytes
                b"\x00\x01\x02",  # Control characters
                b"a" * 1000,  # Long ASCII string
                b"\xff" * 1000,  # Long binary string
            ]
        }
    )
    result = table.eval_expression_list([col("col").binary.length()])
    assert result.to_pydict() == {
        "col": [
            3,  # "foo"
            None,  # None
            0,  # ""
            13,  # "Hello☃World" (5 + 3 + 5 = 13 bytes)
            8,  # "😉test" (4 + 4 = 8 bytes)
            8,  # "test🌈" (4 + 4 = 8 bytes)
            11,  # "☃😉🌈" (3 + 4 + 4 = 11 bytes)
            11,  # "Hello\x00World"
            3,  # "\xff\xfe\xfd"
            3,  # "\x00\x01\x02"
            1000,  # Long ASCII string
            1000,  # Long binary string
        ]
    }


@pytest.mark.parametrize(
    "input_data,expected_lengths",
    [
        # Basic ASCII strings
        ([b"hello", b"world", b"test"], [5, 5, 4]),
        # Empty strings and nulls
        ([b"", None, b"", None], [0, None, 0, None]),
        # Special binary sequences
        (
            [
                b"\x00\x01\x02",  # Control characters
                b"\xff\xfe\xfd",  # High bytes
                b"Hello\x00World",  # String with null
                b"\xe2\x98\x83",  # UTF-8 snowman
                b"\xf0\x9f\x98\x89",  # UTF-8 winking face
            ],
            [3, 3, 11, 3, 4],
        ),
        # Mixed content
        (
            [
                b"Hello\xe2\x98\x83World",  # String with UTF-8
                b"\xf0\x9f\x98\x89test",  # UTF-8 at start
                b"test\xf0\x9f\x8c\x88",  # UTF-8 at end
                b"\xe2\x98\x83\xf0\x9f\x98\x89\xf0\x9f\x8c\x88",  # Multiple UTF-8
            ],
            [13, 8, 8, 11],  # Fixed lengths for UTF-8 sequences
        ),
        # Large strings
        ([b"a" * 1000, b"\xff" * 1000, b"x" * 500], [1000, 1000, 500]),
    ],
)
def test_binary_length_parameterized(input_data: list[bytes | None], expected_lengths: list[int | None]) -> None:
    table = MicroPartition.from_pydict({"col": input_data})
    result = table.eval_expression_list([col("col").binary.length()])
    assert result.to_pydict() == {"col": expected_lengths}


def test_binary_length_errors() -> None:
    # Test length with wrong number of arguments
    table = MicroPartition.from_pydict({"a": [b"hello", b"world"], "b": [b"foo", b"bar"]})
    with pytest.raises(
        Exception, match="(?:ExpressionBinaryNamespace.)?length\\(\\) takes 1 positional argument but 2 were given"
    ):
        table.eval_expression_list([col("a").binary.length(col("b"))])



from __future__ import annotations

import contextlib
import os
import subprocess
import sys

import pytest


@contextlib.contextmanager
def with_null_env():
    old_daft_runner = os.getenv("DAFT_RUNNER")
    del os.environ["DAFT_RUNNER"]

    try:
        yield
    finally:
        if old_daft_runner is not None:
            os.environ["DAFT_RUNNER"] = old_daft_runner


def test_explicit_set_runner_native():
    """Test that a freshly imported context doesn't have a runner config set and can be set explicitly to Native."""
    explicit_set_runner_script_native = """
import daft
print(daft.context.get_context()._runner)
daft.context.set_runner_native()
print(daft.context.get_context()._runner.name)
    """

    with with_null_env():
        result = subprocess.run(
            [sys.executable, "-c", explicit_set_runner_script_native],
            capture_output=True,
        )
        assert result.stdout.decode().strip() == "None\nnative"


def test_implicit_set_runner_native():
    """Test that a freshly imported context doesn't have a runner config set and is set implicitly to Native."""
    implicit_set_runner_script = """
import daft
print(daft.context.get_context()._runner)
df = daft.from_pydict({"foo": [1, 2, 3]})
print(daft.context.get_context()._runner.name)
    """

    with with_null_env():
        result = subprocess.run([sys.executable, "-c", implicit_set_runner_script], capture_output=True)
        assert result.stdout.decode().strip() == "None\nnative"


def test_explicit_set_runner_ray():
    """Test that a freshly imported context doesn't have a runner config set and can be set explicitly to Ray."""
    explicit_set_runner_script_ray = """
import daft
print(daft.context.get_context()._runner)
daft.context.set_runner_ray()
print(daft.context.get_context()._runner.name)
    """

    with with_null_env():
        result = subprocess.run([sys.executable, "-c", explicit_set_runner_script_ray], capture_output=True)
        assert result.stdout.decode().strip() == "None\nray"


def test_implicit_set_runner_ray():
    """Test that a freshly imported context doesn't have a runner config set and can be set implicitly to Ray."""
    implicit_set_runner_script_ray = """
import daft
import ray
ray.init()
print(daft.context.get_context()._runner)
df = daft.from_pydict({"foo": [1, 2, 3]})
print(daft.context.get_context()._runner.name)
    """

    with with_null_env():
        result = subprocess.run([sys.executable, "-c", implicit_set_runner_script_ray], capture_output=True)
        assert result.stdout.decode().strip() == "None\nray"


def test_cannot_switch_local_to_ray():
    """Test that a runner cannot be switched from local to Ray."""
    script = """
import daft
daft.context.set_runner_native()
daft.context.set_runner_ray()
"""
    with with_null_env():
        result = subprocess.run([sys.executable, "-c", script], capture_output=True)
        assert "DaftError::InternalError Cannot set runner more than once" in result.stderr.decode().strip()


@pytest.mark.parametrize(
    "set_new_runner_command",
    [
        pytest.param("daft.context.set_runner_native()", id="ray_to_native"),
        pytest.param("daft.context.set_runner_ray()", id="ray_to_ray"),
    ],
)
def test_cannot_switch_from_ray(set_new_runner_command):
    """Test that a runner cannot be switched from Ray."""
    script = f"""
import daft
daft.context.set_runner_ray()
{set_new_runner_command}
"""
    with with_null_env():
        result = subprocess.run([sys.executable, "-c", script], capture_output=True)
        assert "DaftError::InternalError Cannot set runner more than once" in result.stderr.decode().strip()


@pytest.mark.parametrize("daft_runner_envvar", ["ray", "native"])
def test_env_var(daft_runner_envvar):
    """Test that environment variables are correctly picked up."""
    autodetect_script = """
import daft
df = daft.from_pydict({"foo": [1, 2, 3]})
print(daft.context.get_context()._runner.name)
    """

    with with_null_env():
        result = subprocess.run(
            [sys.executable, "-c", autodetect_script],
            capture_output=True,
            env={"DAFT_RUNNER": daft_runner_envvar},
        )
        assert result.stdout.decode().strip() == daft_runner_envvar


def test_in_ray_job():
    """Test that Ray job ID environment variable is being picked up."""
    autodetect_script = """
import daft
df = daft.from_pydict({"foo": [1, 2, 3]})
print(daft.context.get_context()._runner.name)
    """

    with with_null_env():
        result = subprocess.run(
            [sys.executable, "-c", autodetect_script],
            capture_output=True,
            env={"RAY_JOB_ID": "dummy"},
        )
        assert result.stdout.decode().strip() == "ray"


def test_get_or_create_runner_from_multiple_threads():
    concurrent_get_or_create_runner_script = """
import concurrent.futures
from daft.context import get_context

with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
    futures = [
        executor.submit(lambda: get_context().get_or_create_runner()) for _ in range(10)
    ]

    results = [future.result() for future in concurrent.futures.as_completed(futures)]
    print("ok")
    """

    with with_null_env():
        result = subprocess.run(
            [sys.executable, "-c", concurrent_get_or_create_runner_script],
            capture_output=True,
        )
        assert result.stdout.decode().strip() == "ok"


# TODO: Figure out why these tests are failing for Py3.8
# def test_in_ray_worker():
#     """Test that running Daft in a Ray worker defaults to Python runner"""

#     autodetect_script = """
# import ray
# import os
# import daft

# @ray.remote
# def init_and_return_daft_runner():

#     # Attempt to confuse our runner inference code
#     assert ray.is_initialized()
#     os.environ["RAY_JOB_ID"] = "dummy"

#     df = daft.from_pydict({"foo": [1, 2, 3]})
#     return daft.context.get_context()._runner.name

# ray.init()
# print(ray.get(init_and_return_daft_runner.remote()))
#     """

#     with with_null_env():
#         result = subprocess.run([sys.executable, "-c", autodetect_script], capture_output=True)
#         assert result.stdout.decode().strip() in {"native"}


# def test_in_ray_worker_launch_query():
#     """Test that running Daft in a Ray worker defaults to Python runner"""

#     autodetect_script = """
# import ray
# import os
# import daft

# @ray.remote
# def init_and_return_daft_runner():

#     assert ray.is_initialized()
#     daft.context.set_runner_ray()

#     df = daft.from_pydict({"foo": [1, 2, 3]})
#     return daft.context.get_context()._runner.name

# ray.init()
# print(ray.get(init_and_return_daft_runner.remote()))
#     """

#     with with_null_env():
#         result = subprocess.run([sys.executable, "-c", autodetect_script], capture_output=True)
#         assert result.stdout.decode().strip() == "ray"


def test_cannot_set_runner_ray_after_py():
    cannot_set_runner_ray_after_py_script = """
import daft
df = daft.from_pydict({"foo": [1, 2, 3]})
print(daft.context.get_context()._runner.name)
daft.context.set_runner_ray()
    """

    with with_null_env():
        result = subprocess.run(
            [sys.executable, "-c", cannot_set_runner_ray_after_py_script],
            capture_output=True,
        )
        assert result.stdout.decode().strip() in {"native"}
        assert "DaftError::InternalError Cannot set runner more than once" in result.stderr.decode().strip()



from __future__ import annotations

from daft.expressions import col
from daft.series import Series
from tests.expressions.typing.conftest import (
    assert_typing_resolve_vs_runtime_behavior,
    has_supertype,
)


def test_if_else(binary_data_fixture):
    lhs, rhs = binary_data_fixture
    predicate_series = Series.from_pylist([True, False, None], name="predicate")

    assert_typing_resolve_vs_runtime_behavior(
        data=(*binary_data_fixture, predicate_series),
        expr=col("predicate").if_else(col(lhs.name()), col(rhs.name())),
        run_kernel=lambda: predicate_series.if_else(lhs, rhs),
        resolvable=has_supertype(lhs.datatype(), rhs.datatype()),
    )



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Iterator

    from daft.dataframe import DataFrame
    from daft.io.partitioning import PartitionField
    from daft.io.pushdowns import Pushdowns
    from daft.recordbatch import MicroPartition
    from daft.schema import Schema


__all__ = [
    "DataSource",
    "DataSourceTask",
]


class DataSource(ABC):
    """DataSource is a low-level interface for reading data into DataFrames.

    When a DataSource is read, it is split into multiple tasks which can be distributed
    for parallel processing. Each task is responsible for reading a specific portion of
    the data (e.g., a file partition, a range of rows, or a subset of a database table)
    and converting it into RecordBatches. Implementations should ensure that tasks
    are appropriately sized to balance parallelism.

    Warning:
        This API is early in its development and is subject to change.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """Returns the source name which is useful for debugging."""
        ...

    @property
    @abstractmethod
    def schema(self) -> Schema:
        """Returns the schema shared by each task's record batches."""
        ...

    def get_partition_fields(self) -> list[PartitionField]:
        """Returns the partitioning fields for this data source."""
        return []

    @abstractmethod
    def get_tasks(self, pushdowns: Pushdowns) -> Iterator[DataSourceTask]:
        """Returns an iterator of tasks for this source.

        Returns:
            Iterable[DataSourceTask]: An iterable of tasks that can be processed independently.
        """
        ...

    def read(self) -> DataFrame:
        """Reads a DataSource as a DataFrame."""
        from daft.daft import ScanOperatorHandle
        from daft.dataframe import DataFrame
        from daft.io.__shim import _DataSourceShim
        from daft.logical.builder import LogicalPlanBuilder

        scan = _DataSourceShim(self)
        handle = ScanOperatorHandle.from_python_scan_operator(scan)
        builder = LogicalPlanBuilder.from_tabular_scan(scan_operator=handle)
        return DataFrame(builder)


class DataSourceTask(ABC):
    """DataSourceTask represents a partition of data that can be processed independently.

    Warning:
        This API is early in its development and is subject to change.
    """

    @property
    @abstractmethod
    def schema(self) -> Schema:
        """Returns the schema shared by each MicroPartition."""
        ...

    @abstractmethod
    def get_micro_partitions(self) -> Iterator[MicroPartition]:
        """Executes this task to produce MicroPartitions.

        Returns:
            An iterable of MicroPartition objects containing the data for this task.
        """
        ...



from __future__ import annotations

from daft.expressions import col
from daft.recordbatch import MicroPartition


def test_utf8_substr():
    table = MicroPartition.from_pydict({"col": ["foo", None, "barbarbar", "quux", "1", ""]})
    result = table.eval_expression_list([col("col").str.substr(0, 5)])
    assert result.to_pydict() == {"col": ["foo", None, "barba", "quux", "1", None]}



from __future__ import annotations


def test_into_partitions_some_empty(make_df) -> None:
    data = {"foo": [1, 2, 3]}
    df = make_df(data).into_partitions(32).collect()
    assert df.to_pydict() == data


def test_into_partitions_coalesce(make_df) -> None:
    data = {"foo": list(range(100))}
    df = make_df(data).into_partitions(20).into_partitions(1).collect()
    assert df.to_pydict() == data


def test_into_partitions_some_no_split(make_df) -> None:
    data = {"foo": [1, 2, 3]}

    # Materialize as 3 partitions
    df = make_df(data).into_partitions(3).collect()

    # Attempt to split into 4 partitions, so only 1 split occurs
    df = df.into_partitions(4).collect()

    assert df.to_pydict() == data



from __future__ import annotations

from operator import and_, or_, xor

import pytest

from daft.expressions import col
from daft.recordbatch import MicroPartition

BITWISE_OPERATORS = [and_, or_, xor]


@pytest.mark.parametrize("op", BITWISE_OPERATORS)
@pytest.mark.parametrize(
    "left, right",
    [
        pytest.param([0b1100, 0b1010, 0b1001], [0b1010, 0b1100, 0b1001], id="no_nulls"),
        pytest.param([0b1100, None, 0b1001], [0b1010, 0b1100, None], id="with_nulls"),
        pytest.param([None, None, None], [0b1010, 0b1100, 0b1001], id="left_nulls"),
        pytest.param([0b1100, 0b1010, 0b1001], [None, None, None], id="right_nulls"),
    ],
)
def test_bitwise_op(op, left, right):
    table = MicroPartition.from_pydict({"left": left, "right": right})

    result = table.eval_expression_list([op(col("left"), col("right"))])
    expected = [op(i, j) if i is not None and j is not None else None for i, j in zip(left, right)]
    assert result.to_pydict()["left"] == expected


@pytest.mark.parametrize(
    "expression, op",
    [
        (col("left").bitwise_and(col("right")), and_),
        (col("left").bitwise_or(col("right")), or_),
        (col("left").bitwise_xor(col("right")), xor),
    ],
)
@pytest.mark.parametrize(
    "left, right",
    [
        pytest.param([0b1100, 0b1010, 0b1001], [0b1010, 0b1100, 0b1001], id="no_nulls"),
        pytest.param([0b1100, None, 0b1001], [0b1010, 0b1100, None], id="with_nulls"),
        pytest.param([None, None, None], [0b1010, 0b1100, 0b1001], id="left_nulls"),
        pytest.param([0b1100, 0b1010, 0b1001], [None, None, None], id="right_nulls"),
    ],
)
def test_bitwise_expression(expression, op, left, right):
    table = MicroPartition.from_pydict({"left": left, "right": right})

    result = table.eval_expression_list([expression])
    expected = [op(i, j) if i is not None and j is not None else None for i, j in zip(left, right)]
    assert result.to_pydict()["left"] == expected


@pytest.mark.parametrize("op", BITWISE_OPERATORS)
@pytest.mark.parametrize(
    "data, scalar",
    [
        pytest.param([0b1100, 0b1010, 0b1001], 0b1010, id="no_nulls"),
        pytest.param([0b1100, None, 0b1001], None, id="with_nulls"),
    ],
)
def test_bitwise_scalar(op, data, scalar):
    table = MicroPartition.from_pydict({"data": data})

    result = table.eval_expression_list([op(col("data"), scalar)])
    expected = [op(d, scalar) if d is not None and scalar is not None else None for d in data]
    assert result.to_pydict()["data"] == expected



from __future__ import annotations

import datetime
import itertools
from typing import Callable

import pyarrow as pa
import pytest
import pytz

from daft.datatype import DataType
from daft.expressions import Expression, ExpressionsProjection
from daft.recordbatch import MicroPartition
from daft.series import Series

ALL_DTYPES = [
    (DataType.int8(), pa.array([1, 2, None], type=pa.int8())),
    (DataType.int16(), pa.array([1, 2, None], type=pa.int16())),
    (DataType.int32(), pa.array([1, 2, None], type=pa.int32())),
    (DataType.int64(), pa.array([1, 2, None], type=pa.int64())),
    (DataType.uint8(), pa.array([1, 2, None], type=pa.uint8())),
    (DataType.uint16(), pa.array([1, 2, None], type=pa.uint16())),
    (DataType.uint32(), pa.array([1, 2, None], type=pa.uint32())),
    (DataType.uint64(), pa.array([1, 2, None], type=pa.uint64())),
    (DataType.float32(), pa.array([1, 2, None], type=pa.float32())),
    (DataType.float64(), pa.array([1, 2, None], type=pa.float64())),
    (DataType.string(), pa.array(["1", "2", "3"], type=pa.string())),
    (DataType.bool(), pa.array([True, False, None], type=pa.bool_())),
    (DataType.null(), pa.array([None, None, None], type=pa.null())),
    (DataType.binary(), pa.array([b"1", b"2", None], type=pa.binary())),
    (DataType.fixed_size_binary(1), pa.array([b"1", b"2", None], type=pa.binary(1))),
]

ALL_DATATYPES_BINARY_PAIRS = list(itertools.product(ALL_DTYPES, repeat=2))

ALL_DATATYPES_TERNARY_PAIRS = list(itertools.product(ALL_DTYPES, repeat=3))

ALL_TEMPORAL_DTYPES = [
    (
        DataType.date(),
        pa.array(
            [datetime.date(2021, 1, 1), datetime.date(2021, 1, 2), None],
            type=pa.date32(),
        ),
    ),
    *[
        (
            DataType.timestamp(unit),
            pa.array(
                [datetime.datetime(2021, 1, 1), datetime.datetime(2021, 1, 2), None],
                type=pa.timestamp(unit),
            ),
        )
        for unit in ["ns", "us", "ms"]
    ],
    *[
        (
            DataType.timestamp(unit, "US/Eastern"),
            pa.array(
                [
                    datetime.datetime(2021, 1, 1).astimezone(pytz.timezone("US/Eastern")),
                    datetime.datetime(2021, 1, 2).astimezone(pytz.timezone("US/Eastern")),
                    None,
                ],
                type=pa.timestamp(unit, "US/Eastern"),
            ),
        )
        for unit in ["ns", "us", "ms"]
    ],
    *[
        (
            DataType.timestamp(unit, "Africa/Accra"),
            pa.array(
                [
                    datetime.datetime(2021, 1, 1).astimezone(pytz.timezone("Africa/Accra")),
                    datetime.datetime(2021, 1, 2).astimezone(pytz.timezone("Africa/Accra")),
                    None,
                ],
                type=pa.timestamp(unit, "Africa/Accra"),
            ),
        )
        for unit in ["ns", "us", "ms"]
    ],
]

ALL_DTYPES += ALL_TEMPORAL_DTYPES

ALL_TEMPORAL_DATATYPES_BINARY_PAIRS = [
    ((dt1, data1), (dt2, data2))
    for (dt1, data1), (dt2, data2) in itertools.product(ALL_TEMPORAL_DTYPES, repeat=2)
    if not (
        pa.types.is_timestamp(data1.type)
        and pa.types.is_timestamp(data2.type)
        and (data1.type.tz is None) ^ (data2.type.tz is None)
    )
]

ALL_DATATYPES_BINARY_PAIRS += ALL_TEMPORAL_DATATYPES_BINARY_PAIRS


DECIMAL_DTYPES = [
    (DataType.decimal128(7, 2), pa.array([1, 2, None], type=pa.decimal128(7, 2))),
    (DataType.decimal128(4, 3), pa.array([1, 2, None], type=pa.decimal128(4, 3))),
]

ALL_DATATYPES_TERNARY_PAIRS = [
    ((dt1, data1), (dt2, data2), (dt3, data3))
    for (dt1, data1), (dt2, data2), (dt3, data3) in itertools.product(ALL_DTYPES, repeat=3)
]


@pytest.fixture(
    scope="module",
    params=ALL_DATATYPES_TERNARY_PAIRS,
    ids=[f"{dt1}-{dt2}-{dt3}" for (dt1, _), (dt2, _), (dt3, _) in ALL_DATATYPES_TERNARY_PAIRS],
)
def ternary_data_fixture(request) -> tuple[Series, Series, Series]:
    """Returns ternary permutation of Series' of all DataType pairs."""
    (dt1, data1), (dt2, data2), (dt3, data3) = request.param
    s1 = Series.from_arrow(data1, name="first")
    assert s1.datatype() == dt1
    s2 = Series.from_arrow(data2, name="second")
    assert s2.datatype() == dt2
    s3 = Series.from_arrow(data3, name="third")
    assert s3.datatype() == dt3
    return (s1, s2, s3)


@pytest.fixture(
    scope="module",
    params=ALL_DATATYPES_BINARY_PAIRS,
    ids=[f"{dt1}-{dt2}" for (dt1, _), (dt2, _) in ALL_DATATYPES_BINARY_PAIRS],
)
def binary_data_fixture(request) -> tuple[Series, Series]:
    """Returns binary permutation of Series' of all DataType pairs."""
    (dt1, data1), (dt2, data2) = request.param
    s1 = Series.from_arrow(data1, name="lhs")
    assert s1.datatype() == dt1
    s2 = Series.from_arrow(data2, name="rhs")
    assert s2.datatype() == dt2
    return (s1, s2)


@pytest.fixture(
    scope="module",
    params=ALL_DTYPES,
    ids=[f"{dt}" for (dt, _) in ALL_DTYPES],
)
def unary_data_fixture(request) -> Series:
    """Returns unary permutation of Series' of all DataType pairs."""
    (dt, data) = request.param
    s = Series.from_arrow(data, name="arg")
    assert s.datatype() == dt
    return s


@pytest.fixture(
    scope="module",
    params=DECIMAL_DTYPES,
    ids=[f"{dt}" for (dt, _) in DECIMAL_DTYPES],
)
def decimal_unary_data_fixture(request) -> Series:
    """Returns unary permutation of Series' of select decimal DataType pairs."""
    (dt, data) = request.param
    s = Series.from_arrow(data, name="arg")
    assert s.datatype() == dt
    return s


def assert_typing_resolve_vs_runtime_behavior(
    data: tuple[Series],
    expr: Expression,
    run_kernel: Callable[[], Series],
    resolvable: bool,
):
    """Asserts that typing behavior during schema resolution matches behavior during runtime on Series'.

    Example Usage:

        >>> def my_test(binary_data_fixture):
        >>>     lhs, rhs = binary_data_fixture  # unwrap the generated Series data
        >>>     assert_typing_resolve_vs_runtime_behavior(
        >>>         data=binary_data_fixture,
        >>>         expr=col(lhs.name()) + col(rhs.name()),
        >>>         run_kernel=lambda: lhs + rhs,
        >>>         resolvable=can_add_dtypes(lhs.datatype(), rhs.datatype()),
        >>>     )

    Args:
        data: data to test against (generated using one of the provided fixtures, `{unary, binary}_data_fixture`)
        expr (Expression): Expression used to run the kernel in a MicroPartition (use `.name()` of the generated data to refer to columns)
        run_kernel (Callable): A lambda that will run the kernel directly on the generated Series' without going through the Expressions API
        resolvable (bool): Whether this kernel should be valid, given the datatypes of the generated Series'
    """
    table = MicroPartition.from_pydict({s.name(): s for s in data})
    projection = ExpressionsProjection([expr.alias("result")])
    if resolvable:
        # Check that schema resolution and Series runtime return the same datatype
        resolved_schema = projection.resolve_schema(table.schema())
        result = run_kernel()
        assert (
            resolved_schema["result"].dtype == result.datatype()
        ), "Should have matching result types at runtime and schema-resolve-time"
    else:
        # Check that we fail to resolve types during schema resolution
        with pytest.raises(ValueError):
            projection.resolve_schema(table.schema())
        # TODO: check that types also fail to resolve at runtime
        # with pytest.raises(ValueError):
        #     run_kernel()


def is_numeric_or_null(dt: DataType) -> bool:
    """Checks if this type is a numeric or null type."""
    return dt == DataType.null() or is_numeric(dt)


def is_numeric(dt: DataType) -> bool:
    """Checks if this type is a numeric type."""
    return (
        dt == DataType.int8()
        or dt == DataType.int16()
        or dt == DataType.int32()
        or dt == DataType.int64()
        or dt == DataType.uint8()
        or dt == DataType.uint16()
        or dt == DataType.uint32()
        or dt == DataType.uint64()
        or dt == DataType.float32()
        or dt == DataType.float64()
    )


def is_integer(dt: DataType) -> bool:
    """Checks if this type is an integer type."""
    return (
        dt == DataType.int8()
        or dt == DataType.int16()
        or dt == DataType.int32()
        or dt == DataType.int64()
        or dt == DataType.uint8()
        or dt == DataType.uint16()
        or dt == DataType.uint32()
        or dt == DataType.uint64()
    )


def is_signed_integer(dt: DataType) -> bool:
    """Checks if this type is a signed integer type."""
    return dt == DataType.int8() or dt == DataType.int16() or dt == DataType.int32() or dt == DataType.int64()


def is_comparable(dt: DataType):
    """Checks if this type is a comparable type."""
    return (
        is_numeric(dt)
        or dt == DataType.bool()
        or dt == DataType.string()
        or dt == DataType.null()
        or dt == DataType.binary()
        or dt == DataType.fixed_size_binary(1)
        or dt.is_temporal()
    )


def is_numeric_bitwidth_gte_32(dt: DataType):
    """Checks if type is numeric and above a bitwidth of 32."""
    return (
        dt == DataType.int32()
        or dt == DataType.int64()
        or dt == DataType.uint32()
        or dt == DataType.uint64()
        or dt == DataType.float32()
        or dt == DataType.float64()
    )


def has_supertype(dt1: DataType, dt2: DataType) -> bool:
    """Checks if two DataTypes have supertypes.

    this is a simplified
    version of `supertype.rs`, since it only defines "reachability" within the supertype
    tree in a more human-readable way for testing purposes.
    """
    # super(T, T) = T
    if dt1 == dt2:
        return True

    for x, y in ((dt1, dt2), (dt2, dt1)):
        # --- Common types across hierarchies ---
        either_null = x == DataType.null()
        either_string_and_other_not_binary = x == DataType.string() and not (
            y == DataType.binary() or y == DataType.fixed_size_binary(1)
        )

        # --- Within type hierarchies ---
        both_numeric = (is_numeric(x) and is_numeric(y)) or ((x == DataType.bool()) and is_numeric(y))
        both_temporal = x.is_temporal() and y.is_temporal()

        # --- Across type hierarchies ---
        date_and_numeric = x == DataType.date() and is_numeric(y)
        time_and_numeric = x == (DataType.time("us") or DataType.time("ns")) and is_numeric(y)
        timestamp_and_big_numeric = x.is_temporal() and is_numeric_bitwidth_gte_32(y)

        if (
            either_null
            or either_string_and_other_not_binary
            or both_numeric
            or both_temporal
            or date_and_numeric
            or time_and_numeric
            or timestamp_and_big_numeric
        ):
            return True

    return False



from __future__ import annotations

import pytest

import daft
from daft.datatype import DataType


@pytest.fixture(scope="function")
def df():
    return daft.from_pydict({"foo": [1, 2, 3]})


def test_num_partitions(df):
    assert df.num_partitions() == 1

    df2 = df.repartition(2)
    assert df2.num_partitions() == 2


def test_schema(df):
    fields = [f for f in df.schema()]
    assert len(fields) == 1
    [field] = fields
    assert field.name == "foo"
    assert field.dtype == DataType.int64()


def test_column_names(df):
    assert df.column_names == ["foo"]


def test_columns(df):
    assert len(df.columns) == 1
    [ex] = df.columns
    assert ex.name() == "foo"


def test_in(df):
    assert "foo" in df
    assert "bar" not in df



from __future__ import annotations

import argparse
import logging
import math
import os
import shlex
import sqlite3
import subprocess
from glob import glob

import daft

logger = logging.getLogger(__name__)

SCHEMA = {
    "part": [
        "P_PARTKEY",
        "P_NAME",
        "P_MFGR",
        "P_BRAND",
        "P_TYPE",
        "P_SIZE",
        "P_CONTAINER",
        "P_RETAILPRICE",
        "P_COMMENT",
    ],
    "supplier": [
        "S_SUPPKEY",
        "S_NAME",
        "S_ADDRESS",
        "S_NATIONKEY",
        "S_PHONE",
        "S_ACCTBAL",
        "S_COMMENT",
    ],
    "partsupp": [
        "PS_PARTKEY",
        "PS_SUPPKEY",
        "PS_AVAILQTY",
        "PS_SUPPLYCOST",
        "PS_COMMENT",
    ],
    "customer": [
        "C_CUSTKEY",
        "C_NAME",
        "C_ADDRESS",
        "C_NATIONKEY",
        "C_PHONE",
        "C_ACCTBAL",
        "C_MKTSEGMENT",
        "C_COMMENT",
    ],
    "orders": [
        "O_ORDERKEY",
        "O_CUSTKEY",
        "O_ORDERSTATUS",
        "O_TOTALPRICE",
        "O_ORDERDATE",
        "O_ORDERPRIORITY",
        "O_CLERK",
        "O_SHIPPRIORITY",
        "O_COMMENT",
    ],
    "lineitem": [
        "L_ORDERKEY",
        "L_PARTKEY",
        "L_SUPPKEY",
        "L_LINENUMBER",
        "L_QUANTITY",
        "L_EXTENDEDPRICE",
        "L_DISCOUNT",
        "L_TAX",
        "L_RETURNFLAG",
        "L_LINESTATUS",
        "L_SHIPDATE",
        "L_COMMITDATE",
        "L_RECEIPTDATE",
        "L_SHIPINSTRUCT",
        "L_SHIPMODE",
        "L_COMMENT",
    ],
    "nation": [
        "N_NATIONKEY",
        "N_NAME",
        "N_REGIONKEY",
        "N_COMMENT",
    ],
    "region": [
        "R_REGIONKEY",
        "R_NAME",
        "R_COMMENT",
    ],
}


TPCH_SQLITE_TABLE_CREATION_SQL = [
    """
        CREATE TABLE NATION (
        N_NATIONKEY INTEGER PRIMARY KEY NOT NULL,
        N_NAME      TEXT NOT NULL,
        N_REGIONKEY INTEGER NOT NULL,
        N_COMMENT   TEXT,
        FOREIGN KEY (N_REGIONKEY) REFERENCES REGION(R_REGIONKEY)
        );
    """,
    """
        CREATE TABLE REGION (
        R_REGIONKEY INTEGER PRIMARY KEY NOT NULL,
        R_NAME      TEXT NOT NULL,
        R_COMMENT   TEXT
        );
    """,
    """
        CREATE TABLE PART (
        P_PARTKEY     INTEGER PRIMARY KEY NOT NULL,
        P_NAME        TEXT NOT NULL,
        P_MFGR        TEXT NOT NULL,
        P_BRAND       TEXT NOT NULL,
        P_TYPE        TEXT NOT NULL,
        P_SIZE        INTEGER NOT NULL,
        P_CONTAINER   TEXT NOT NULL,
        P_RETAILPRICE REAL NOT NULL,
        P_COMMENT     TEXT NOT NULL
        );
    """,
    """
        CREATE TABLE SUPPLIER (
        S_SUPPKEY   INTEGER PRIMARY KEY NOT NULL,
        S_NAME      TEXT NOT NULL,
        S_ADDRESS   TEXT NOT NULL,
        S_NATIONKEY INTEGER NOT NULL,
        S_PHONE     TEXT NOT NULL,
        S_ACCTBAL   REAL NOT NULL,
        S_COMMENT   TEXT NOT NULL,
        FOREIGN KEY (S_NATIONKEY) REFERENCES NATION(N_NATIONKEY)
        );
    """,
    """
        CREATE TABLE PARTSUPP (
        PS_PARTKEY    INTEGER NOT NULL,
        PS_SUPPKEY    INTEGER NOT NULL,
        PS_AVAILQTY   INTEGER NOT NULL,
        PS_SUPPLYCOST REAL NOT NULL,
        PS_COMMENT    TEXT NOT NULL,
        PRIMARY KEY (PS_PARTKEY, PS_SUPPKEY),
        FOREIGN KEY (PS_SUPPKEY) REFERENCES SUPPLIER(S_SUPPKEY),
        FOREIGN KEY (PS_PARTKEY) REFERENCES PART(P_PARTKEY)
        );
    """,
    """
        CREATE TABLE CUSTOMER (
        C_CUSTKEY    INTEGER PRIMARY KEY NOT NULL,
        C_NAME       TEXT NOT NULL,
        C_ADDRESS    TEXT NOT NULL,
        C_NATIONKEY  INTEGER NOT NULL,
        C_PHONE      TEXT NOT NULL,
        C_ACCTBAL    REAL   NOT NULL,
        C_MKTSEGMENT TEXT NOT NULL,
        C_COMMENT    TEXT NOT NULL,
        FOREIGN KEY (C_NATIONKEY) REFERENCES NATION(N_NATIONKEY)
        );
    """,
    """
        CREATE TABLE ORDERS (
        O_ORDERKEY      INTEGER PRIMARY KEY NOT NULL,
        O_CUSTKEY       INTEGER NOT NULL,
        O_ORDERSTATUS   TEXT NOT NULL,
        O_TOTALPRICE    REAL NOT NULL,
        O_ORDERDATE     DATE NOT NULL,
        O_ORDERPRIORITY TEXT NOT NULL,
        O_CLERK         TEXT NOT NULL,
        O_SHIPPRIORITY  INTEGER NOT NULL,
        O_COMMENT       TEXT NOT NULL,
        FOREIGN KEY (O_CUSTKEY) REFERENCES CUSTOMER(C_CUSTKEY)
        );
    """,
    """
        CREATE TABLE LINEITEM (
        L_ORDERKEY      INTEGER NOT NULL,
        L_PARTKEY       INTEGER NOT NULL,
        L_SUPPKEY       INTEGER NOT NULL,
        L_LINENUMBER    INTEGER NOT NULL,
        L_QUANTITY      REAL NOT NULL,
        L_EXTENDEDPRICE REAL NOT NULL,
        L_DISCOUNT      REAL NOT NULL,
        L_TAX           REAL NOT NULL,
        L_RETURNFLAG    TEXT NOT NULL,
        L_LINESTATUS    TEXT NOT NULL,
        L_SHIPDATE      DATE NOT NULL,
        L_COMMITDATE    DATE NOT NULL,
        L_RECEIPTDATE   DATE NOT NULL,
        L_SHIPINSTRUCT  TEXT NOT NULL,
        L_SHIPMODE      TEXT NOT NULL,
        L_COMMENT       TEXT NOT NULL,
        PRIMARY KEY (L_ORDERKEY, L_LINENUMBER),
        FOREIGN KEY (L_ORDERKEY) REFERENCES ORDERS(O_ORDERKEY),
        FOREIGN KEY (L_PARTKEY, L_SUPPKEY) REFERENCES PARTSUPP(PS_PARTKEY, PS_SUPPKEY)
        );
    """,
]


def gen_sqlite_db(csv_filepath: str, num_parts: int) -> str:
    """Generates a SQLite DB from a folder filled with generated CSVs.

    Args:
        csv_filepath (str): path to folder with generated CSVs
        sqlite_db_file_path (str): path to write sqlite db to
        num_parts (int): number of parts of CSVs
    """
    sqlite_db_file_path = os.path.join(csv_filepath, "TPC-H.db")

    def import_table(table, table_path):
        if not os.path.isfile(table_path):
            raise FileNotFoundError(f"Did not find {table_path}")
        subprocess.check_output(
            shlex.split(f"sqlite3 {sqlite_db_file_path} -cmd '.mode csv' '.separator |' '.import {table_path} {table}'")
        )

    if not os.path.exists(sqlite_db_file_path):
        logger.info("Generating SQLite DB at %s", sqlite_db_file_path)
        con = sqlite3.connect(sqlite_db_file_path)
        cur = con.cursor()
        for creation_sql in TPCH_SQLITE_TABLE_CREATION_SQL:
            cur.execute(creation_sql)

        static_tables = ["nation", "region"]
        partitioned_tables = ["customer", "lineitem", "orders", "partsupp", "part", "supplier"]
        single_partition_tables = static_tables + partitioned_tables if num_parts == 1 else static_tables
        multi_partition_tables = [] if num_parts == 1 else partitioned_tables

        for table in single_partition_tables:
            import_table(table, f"{csv_filepath}/{table}.tbl")

        for table in multi_partition_tables:
            for part_idx in range(num_parts):
                import_table(table, f"{csv_filepath}/{table}.tbl.{part_idx}")
    else:
        logger.info("Cached SQLite DB already exists at %s, skipping SQLite DB generation", sqlite_db_file_path)

    return sqlite_db_file_path


def gen_csv_files(basedir: str, num_parts: int, scale_factor: float) -> str:
    """Generates CSV files.

    Args:
        basedir (str): path to generate files into
        num_parts (int): number of parts to generate
        scale_factor (float): scale factor for generation

    Returns:
        str: path to folder with generated CSV files
    """
    cachedir = os.path.join(basedir, (f"{scale_factor:.1f}").replace(".", "_"), str(num_parts))
    if not os.path.exists(cachedir):
        # If running in CI, use a scale factor of 0.2
        # Otherwise, check for SCALE_FACTOR env variable or default to 1
        logger.info("Cloning tpch dbgen repo")
        subprocess.check_output(shlex.split(f"git clone https://github.com/electrum/tpch-dbgen {cachedir}"))
        subprocess.check_output("make", cwd=cachedir)

        logger.info("Generating %s for tpch", num_parts)
        subprocess.check_output(shlex.split(f"./dbgen -v -f -s {scale_factor}"), cwd=cachedir)

        # The tool sometimes generates weird files with bad permissioning, we override that manually here
        subprocess.check_output(shlex.split("chmod -R u+rwx ."), cwd=cachedir)

        # Split the files manually
        if num_parts > 1:
            for tbl in ["customer", "lineitem", "orders", "partsupp", "part", "supplier"]:
                csv_file = os.path.join(cachedir, f"{tbl}.tbl")
                with open(csv_file) as f:
                    num_lines = sum(1 for _ in f)
                num_lines_per_part = math.ceil(num_lines / num_parts)
                with open(csv_file, "rb") as f:
                    for i in range(num_parts):
                        csv_part_file = os.path.join(cachedir, f"{tbl}.tbl.{i}")
                        with open(csv_part_file, "wb") as out:
                            for _ in range(num_lines_per_part):
                                out.write(f.readline())
                os.remove(csv_file)

        # The tool generates CSV files with a trailing delimiter, we get rid of that here
        csv_files = glob(f"{cachedir}/*.tbl*")
        for csv_file in csv_files:
            subprocess.check_output(shlex.split(f"sed -e 's/|$//' -i.bak {csv_file}"))
        backup_files = glob(f"{cachedir}/*.bak")
        for backup_file in backup_files:
            os.remove(backup_file)

    else:
        logger.info(
            "Cached directory for scale_factor=%s num_parts=%s already exists at %s, skipping CSV file generation",
            scale_factor,
            num_parts,
            cachedir,
        )

    return cachedir


def gen_parquet(csv_files_location: str) -> str:
    """Generates Parquet from generated CSV files.

    Args:
        csv_files_location (str): path to folder with generated CSV files

    Returns:
        str: path to generated Parquet files
    """
    PARQUET_FILE_PATH = os.path.join(csv_files_location, "parquet")
    for tab_name in SCHEMA.keys():
        table_parquet_path = os.path.join(PARQUET_FILE_PATH, tab_name)
        if not os.path.exists(table_parquet_path):
            logger.info("Generating Parquet Files for %s", tab_name)
            files = glob(os.path.join(csv_files_location, f"{tab_name}.tbl*"))
            for file in files:
                df = daft.read_csv(
                    file,
                    has_headers=False,
                    delimiter="|",
                ).exclude("")
                df = df.select(
                    *[
                        df[autogen_col_name].alias(actual_col_name)
                        for actual_col_name, autogen_col_name in zip(SCHEMA[tab_name], df.column_names)
                    ]
                )
                df = df.write_parquet(table_parquet_path)
                df.collect()
                assert os.path.exists(
                    table_parquet_path
                ), f"Parquet files not generated by Daft at {table_parquet_path}"
        else:
            logger.info(
                "Cached Parquet files for table %s already exists at %s, skipping Parquet file generation",
                tab_name,
                table_parquet_path,
            )
    return PARQUET_FILE_PATH


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--tpch_gen_folder",
        default="data/tpch-dbgen",
        help="Path to the folder containing the TPCH dbgen tool and generated data",
    )
    parser.add_argument("--scale_factor", default=10.0, help="Scale factor to run on in GB", type=float)
    parser.add_argument(
        "--num_parts", default=None, help="Number of parts to generate (defaults to 1 part per GB)", type=int
    )
    parser.add_argument(
        "--generate_sqlite_db", action="store_true", help="Whether to generate a SQLite DB or not, defaults to False"
    )
    parser.add_argument(
        "--generate_parquet", action="store_true", help="Whether to generate Parquet files or not, defaults to False"
    )
    args = parser.parse_args()
    num_parts = math.ceil(args.scale_factor) if args.num_parts is None else args.num_parts

    logger.info(
        "Generating data at %s with: scale_factor=%s num_parts=%s generate_sqlite_db=%s generate_parquet=%s",
        args.tpch_gen_folder,
        args.scale_factor,
        num_parts,
        args.generate_sqlite_db,
        args.generate_parquet,
    )

    csv_folder = gen_csv_files(basedir=args.tpch_gen_folder, scale_factor=args.scale_factor, num_parts=num_parts)
    if args.generate_parquet:
        gen_parquet(csv_folder)
    if args.generate_sqlite_db:
        gen_sqlite_db(csv_folder, num_parts)



from __future__ import annotations

import os
from dataclasses import dataclass

from daft.dependencies import pa, pafs, pq
from daft.filesystem import join_path


@dataclass(init=False)
class FsFileMetadata:
    def __init__(self, fs: pafs.FileSystem, base_path: str, path: str, base_name: str):
        self.base_path = base_path
        self.path = path
        self.base_name = base_name
        with fs.open_input_file(join_path(fs, base_path, path)) as f:
            metadata = pq.read_metadata(f)
        self.size = metadata.serialized_size
        self.num_records = metadata.num_rows
        self.schema = FsFileMetadata._extract_arrow_schema(metadata)
        self.colstats_schema, self.min_values, self.max_values = FsFileMetadata._extract_min_max(metadata)

    @staticmethod
    def _get_colstats_schema(original_schema: pa.Schema, extracted_field_names: set[str]) -> pa.Schema:
        extracted_fields = []
        for i in range(len(original_schema)):
            f = original_schema.field(i)
            if f.name in extracted_field_names:
                extracted_fields.append(f)
        return pa.schema(extracted_fields)

    @staticmethod
    def _extract_arrow_schema(metadata: pq.FileMetaData) -> pa.Schema:
        return metadata.schema.to_arrow_schema().remove_metadata()

    @staticmethod
    def _is_nested(path_in_schema: str) -> bool:
        return path_in_schema.find(".") != -1

    @staticmethod
    def _extract_min_max(metadata: pq.FileMetaData) -> tuple[pa.Schema, list[int | None], list[int | None]]:
        num_columns = metadata.num_columns
        num_row_groups = metadata.num_row_groups
        min_vals: list[int | None] = [None] * num_columns
        max_vals: list[int | None] = [None] * num_columns
        colstats_field_names = set()
        for rg in range(num_row_groups):
            row_group = metadata.row_group(rg)
            for col in range(num_columns):
                column = row_group.column(col)
                if (
                    column.is_stats_set
                    and column.statistics is not None
                    and column.statistics.has_min_max
                    and not FsFileMetadata._is_nested(column.path_in_schema)
                ):
                    stats = column.statistics
                    assert stats is not None
                    colstats_field_names.add(column.path_in_schema)
                    if min_vals[col] is None or (stats.min and stats.min < min_vals[col]):
                        min_vals[col] = stats.min
                    if max_vals[col] is None or (stats.max and stats.max > max_vals[col]):
                        max_vals[col] = stats.max
        filtered_min_vals = list(filter(lambda v: v is not None, min_vals))
        filtered_max_vals = list(filter(lambda v: v is not None, max_vals))

        colstats_schema = FsFileMetadata._get_colstats_schema(metadata.schema.to_arrow_schema(), colstats_field_names)
        assert len(colstats_schema) == len(filtered_min_vals) == len(filtered_max_vals)
        return colstats_schema, filtered_min_vals, filtered_max_vals


def list_relative_file_paths(
    base_path: str, sub_path: str, fs: pafs.FileSystem, includes: list[str] | None
) -> list[FsFileMetadata]:
    listed_paths: list[pafs.FileInfo] = fs.get_file_info(pafs.FileSelector(join_path(fs, base_path, sub_path)))
    file_paths = []
    common_prefix_len = len(base_path) + 1
    for listed_path in listed_paths:
        if listed_path.type == pafs.FileType.File:
            if includes and os.path.splitext(listed_path.base_name)[-1] in includes:
                file_paths.append(
                    FsFileMetadata(fs, base_path, listed_path.path[common_prefix_len:], listed_path.base_name)
                )

    return file_paths


def list_full_sub_dirs(path: str, fs: pafs.FileSystem, excludes: list[str] | None) -> list[str]:
    sub_paths: list[pafs.FileInfo] = fs.get_file_info(pafs.FileSelector(path))
    sub_dirs = []
    for sub_path in sub_paths:
        if sub_path.type == pafs.FileType.Directory:
            if not excludes or (excludes and sub_path.base_name not in excludes):
                sub_dirs.append(sub_path.path)

    return sub_dirs


def list_leaf_dirs(path: str, fs: pafs.FileSystem) -> list[str]:
    sub_paths: list[pafs.FileInfo] = fs.get_file_info(pafs.FileSelector(path))
    leaf_dirs = []

    for sub_path in sub_paths:
        if sub_path.type == pafs.FileType.Directory:
            leaf_dirs.extend(list_leaf_dirs(sub_path.path, fs))

    # leaf directory
    if len(leaf_dirs) == 0:
        leaf_dirs.append(path)

    return leaf_dirs



from __future__ import annotations

import pyarrow as pa
import pytest

from daft.datatype import DataType
from daft.expressions import col
from daft.recordbatch import MicroPartition


def test_list_slice_empty_series():
    table = MicroPartition.from_pydict(
        {
            "col": pa.array([], type=pa.list_(pa.int64())),
            "start": pa.array([], type=pa.int64()),
            "end": pa.array([], type=pa.int64()),
        }
    )

    result = table.eval_expression_list(
        [
            col("col").list.slice(0, 1).alias("col"),
            col("col").list.slice(0).alias("col-noend"),
            col("col").list.slice(col("start"), 1).alias("col-start"),
            col("col").list.slice(col("start")).alias("col-start-noend"),
            col("col").list.slice(col("start"), col("end")).alias("col-start-end"),
            col("col").list.slice(0, col("end")).alias("col-end"),
        ]
    )

    assert result.to_pydict() == {
        "col": [],
        "col-noend": [],
        "col-start": [],
        "col-start-noend": [],
        "col-start-end": [],
        "col-end": [],
    }


def test_list_slice():
    table = MicroPartition.from_pydict(
        {
            # Test list of an atomic type.
            "col1": [["a"], ["ab", "a"], [None, "a", "", "b", "c"], None, ["a", ""]],
            # Test lists of a nested type.
            "col2": [
                [[1]],
                [[3, 3], [4], [5, 5]],
                [],
                [[], []],
                None,
            ],
            "start": [-1, 1, 0, 2, -2],
            "end": [1, 2, 0, 4, 3],
            "edge_start": [-1, -2, -5, 0, -2],
            "edge_end": [-1, -2, -1, -2, -1],
        }
    )

    result = table.eval_expression_list(
        [
            col("col1").list.slice(0, 1).alias("col1"),
            col("col1").list.slice(0).alias("col1-noend"),
            col("col1").list.slice(col("start"), 1).alias("col1-start"),
            col("col1").list.slice(col("start")).alias("col1-start-noend"),
            col("col1").list.slice(col("start"), col("end")).alias("col1-start-end"),
            col("col1").list.slice(1, col("end")).alias("col1-end"),
            col("col1").list.slice(20, 25).alias("col1-invalid-start"),
            col("col1").list.slice(20, 25).alias("col1-invalid-start-noend"),
            col("col2").list.slice(0, 1).alias("col2"),
            col("col2").list.slice(0).alias("col2-noend"),
            col("col2").list.slice(col("start"), 1).alias("col2-start"),
            col("col2").list.slice(col("start")).alias("col2-start-noend"),
            col("col2").list.slice(col("start"), col("end")).alias("col2-start-end"),
            col("col2").list.slice(0, col("end")).alias("col2-end"),
            col("col2").list.slice(20, 25).alias("col2-invalid-start"),
            col("col2").list.slice(20, 25).alias("col2-invalid-start-noend"),
            # Test edge cases.
            col("col1").list.slice(-10, -20).alias("col1-edge1"),
            col("col1").list.slice(-20, -10).alias("col1-edge2"),
            col("col1").list.slice(-20, 10).alias("col1-edge3"),
            col("col1").list.slice(-20, -1).alias("col1-edge4"),
            col("col1").list.slice(col("edge_start"), col("edge_end")).alias("col1-edge5"),
            col("col1").list.slice(10, 1).alias("col1-edge6"),
            col("col1").list.slice(1, -1).alias("col1-edge7"),
        ]
    )

    assert result.to_pydict() == {
        "col1": [["a"], ["ab"], [None], None, ["a"]],
        "col1-noend": [["a"], ["ab", "a"], [None, "a", "", "b", "c"], None, ["a", ""]],
        "col1-start": [["a"], [], [None], None, ["a"]],
        "col1-start-noend": [["a"], ["a"], [None, "a", "", "b", "c"], None, ["a", ""]],
        "col1-start-end": [["a"], ["a"], [], None, ["a", ""]],
        "col1-end": [[], ["a"], [], None, [""]],
        "col1-invalid-start": [[], [], [], None, []],
        "col1-invalid-start-noend": [[], [], [], None, []],
        "col2": [[[1]], [[3, 3]], [], [[]], None],
        "col2-noend": [[[1]], [[3, 3], [4], [5, 5]], [], [[], []], None],
        "col2-start": [[[1]], [], [], [], None],
        "col2-start-noend": [[[1]], [[4], [5, 5]], [], [], None],
        "col2-start-end": [[[1]], [[4]], [], [], None],
        "col2-end": [[[1]], [[3, 3], [4]], [], [[], []], None],
        "col2-invalid-start": [[], [], [], [], None],
        "col2-invalid-start-noend": [[], [], [], [], None],
        "col1-edge1": [[], [], [], None, []],
        "col1-edge2": [[], [], [], None, []],
        "col1-edge3": [["a"], ["ab", "a"], [None, "a", "", "b", "c"], None, ["a", ""]],
        "col1-edge4": [[], ["ab"], [None, "a", "", "b"], None, ["a"]],
        "col1-edge5": [[], [], [None, "a", "", "b"], None, ["a"]],
        "col1-edge6": [[], [], [], None, []],
        "col1-edge7": [[], [], ["a", "", "b"], None, []],
    }


def test_fixed_size_list_slice():
    table = MicroPartition.from_pydict(
        {
            # Test list of an atomic type.
            "col1": [["a", "b"], ["aa", "bb"], None, [None, "bbbb"], ["aaaaa", None]],
            # Test lists of a nested type.
            "col2": [
                [[1], [2]],
                [[11, 111], [22, 222]],
                None,
                [None, [3333]],
                [[], []],
            ],
            "start": [-1, 1, 0, 2, -2],
            "end": [1, 1, 0, 2, 3],
            "edge_start": [-1, -2, -5, 0, -2],
            "edge_end": [-1, -2, -1, -2, -1],
        }
    )

    dtype1 = DataType.fixed_size_list(DataType.string(), 2)
    dtype2 = DataType.fixed_size_list(DataType.list(DataType.int32()), 2)

    table = table.eval_expression_list(
        [
            col("col1").cast(dtype1),
            col("col2").cast(dtype2),
            col("start"),
            col("end"),
            col("edge_start"),
            col("edge_end"),
        ]
    )

    result = table.eval_expression_list(
        [
            col("col1").list.slice(0, 1).alias("col1"),
            col("col1").list.slice(0).alias("col1-noend"),
            col("col1").list.slice(col("start"), 1).alias("col1-start"),
            col("col1").list.slice(col("start")).alias("col1-start-noend"),
            col("col1").list.slice(col("start"), col("end")).alias("col1-start-end"),
            col("col1").list.slice(1, col("end")).alias("col1-end"),
            col("col1").list.slice(20, 25).alias("col1-invalid-start"),
            col("col1").list.slice(20).alias("col1-invalid-start-noend"),
            col("col2").list.slice(0, 1).alias("col2"),
            col("col2").list.slice(0).alias("col2-noend"),
            col("col2").list.slice(col("start"), 2).alias("col2-start"),
            col("col2").list.slice(col("start")).alias("col2-start-noend"),
            col("col2").list.slice(col("start"), col("end")).alias("col2-start-end"),
            col("col2").list.slice(0, col("end")).alias("col2-end"),
            col("col2").list.slice(20, 25).alias("col2-invalid-start"),
            col("col2").list.slice(20, 25).alias("col2-invalid-start-noend"),
            # Test edge cases.
            col("col1").list.slice(-10, -20).alias("col1-edge1"),
            col("col1").list.slice(-20, -10).alias("col1-edge2"),
            col("col1").list.slice(-20, 10).alias("col1-edge3"),
            col("col1").list.slice(-20, -1).alias("col1-edge4"),
            col("col1").list.slice(col("edge_start"), col("edge_end")).alias("col1-edge5"),
            col("col1").list.slice(10, 1).alias("col1-edge6"),
            col("col1").list.slice(0, -1).alias("col1-edge7"),
        ]
    )

    assert result.to_pydict() == {
        "col1": [["a"], ["aa"], None, [None], ["aaaaa"]],
        "col1-noend": [["a", "b"], ["aa", "bb"], None, [None, "bbbb"], ["aaaaa", None]],
        "col1-start": [[], [], None, [], ["aaaaa"]],
        "col1-start-noend": [["b"], ["bb"], None, [], ["aaaaa", None]],
        "col1-start-end": [[], [], None, [], ["aaaaa", None]],
        "col1-end": [[], [], None, ["bbbb"], [None]],
        "col1-invalid-start": [[], [], None, [], []],
        "col1-invalid-start-noend": [[], [], None, [], []],
        "col2": [[[1]], [[11, 111]], None, [None], [[]]],
        "col2-noend": [[[1], [2]], [[11, 111], [22, 222]], None, [None, [3333]], [[], []]],
        "col2-start": [[[2]], [[22, 222]], None, [], [[], []]],
        "col2-start-noend": [[[2]], [[22, 222]], None, [], [[], []]],
        "col2-start-end": [[], [], None, [], [[], []]],
        "col2-end": [[[1]], [[11, 111]], None, [None, [3333]], [[], []]],
        "col2-invalid-start": [[], [], None, [], []],
        "col2-invalid-start-noend": [[], [], None, [], []],
        "col1-edge1": [[], [], None, [], []],
        "col1-edge2": [[], [], None, [], []],
        "col1-edge3": [["a", "b"], ["aa", "bb"], None, [None, "bbbb"], ["aaaaa", None]],
        "col1-edge4": [["a"], ["aa"], None, [None], ["aaaaa"]],
        "col1-edge5": [[], [], None, [], ["aaaaa"]],
        "col1-edge6": [[], [], None, [], []],
        "col1-edge7": [["a"], ["aa"], None, [None], ["aaaaa"]],
    }


def test_list_slice_invalid_parameters():
    table = MicroPartition.from_pydict(
        {
            "col": [["a", "b", "c"], ["aa", "bb", "cc"], None, [None, "bbbb"], ["aaaaa", None]],
            "start": [0, -1, 1, 3, -4],
            "end": [1, 2, 3, -1, 0],
        }
    )
    with pytest.raises(ValueError, match="Start index must be an integer, received: Float64"):
        table.eval_expression_list([col("col").list.slice(1.0, 0)])
    with pytest.raises(ValueError, match="End index must be an integer, received: Float64"):
        table.eval_expression_list([col("col").list.slice(0, 1.0)])
    with pytest.raises(TypeError, match="missing 1 required positional argument: 'start'"):
        table.eval_expression_list([col("col").list.slice()])
    with pytest.raises(TypeError, match="takes from 2 to 3 positional arguments but 4 were given"):
        table.eval_expression_list([col("col").list.slice(0, 0, 0)])


def test_list_slice_non_list_type():
    table = MicroPartition.from_pydict(
        {
            "structcol": [{"a": 1}, {"b": 1}, {"c": 1}],
            "stringcol": ["a", "b", "c"],
            "intcol": [1, 2, 3],
        },
    )

    with pytest.raises(ValueError):
        table.eval_expression_list([col("structcol").list.slice(0, 2)])
    with pytest.raises(ValueError):
        table.eval_expression_list([col("stringcol").list.slice(0, 2)])
    with pytest.raises(ValueError):
        table.eval_expression_list([col("intcol").list.slice(0, 2)])



from __future__ import annotations

import datetime

import pytest

from daft.datatype import DataType
from daft.expressions.expressions import col
from daft.recordbatch.micropartition import MicroPartition


@pytest.mark.parametrize(
    "input,fill_value,expected",
    [
        pytest.param([None, None, None], "a", ["a", "a", "a"], id="NullColumn"),
        pytest.param([True, False, None], False, [True, False, False], id="BoolColumn"),
        pytest.param(["a", "b", None], "b", ["a", "b", "b"], id="StringColumn"),
        pytest.param([b"a", None, b"c"], b"b", [b"a", b"b", b"c"], id="BinaryColumn"),
        pytest.param([-1, None, 3], 0, [-1, 0, 3], id="IntColumn"),
        pytest.param([-1.0, None, 3.0], 0.0, [-1.0, 0.0, 3.0], id="FloatColumn"),
        pytest.param(
            [datetime.date.today(), None, datetime.date(2023, 1, 1)],
            datetime.date(2022, 1, 1),
            [
                datetime.date.today(),
                datetime.date(2022, 1, 1),
                datetime.date(2023, 1, 1),
            ],
        ),
        pytest.param(
            [datetime.datetime(2022, 1, 1), None, datetime.datetime(2023, 1, 1)],
            datetime.datetime(2022, 1, 1),
            [
                datetime.datetime(2022, 1, 1),
                datetime.datetime(2022, 1, 1),
                datetime.datetime(2023, 1, 1),
            ],
        ),
    ],
)
def test_table_expr_fill_null(input, fill_value, expected) -> None:
    daft_recordbatch = MicroPartition.from_pydict({"input": input})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("input").fill_null(fill_value)])
    pydict = daft_recordbatch.to_pydict()

    assert pydict["input"] == expected


@pytest.mark.parametrize(
    "float_dtype",
    [DataType.float32(), DataType.float64()],
)
def test_table_expr_fill_nan(float_dtype) -> None:
    input = [1.0, None, 3.0, float("nan")]
    fill_value = 2.0
    expected = [1.0, None, 3.0, 2.0]

    daft_recordbatch = MicroPartition.from_pydict({"input": input})
    daft_recordbatch = daft_recordbatch.eval_expression_list(
        [col("input").cast(float_dtype).float.fill_nan(fill_value)]
    )
    pydict = daft_recordbatch.to_pydict()

    assert pydict["input"] == expected



from __future__ import annotations

from datetime import timedelta


def test_total_seconds(test_expression):
    test_data = [
        timedelta(seconds=1),
        timedelta(milliseconds=1),
        timedelta(microseconds=1),
        timedelta(days=1),
        timedelta(hours=1),
        timedelta(minutes=1),
    ]
    expected = [1, 0, 0, 86400, 3600, 60]
    test_expression(data=test_data, expected=expected, name="total_seconds", sql_name="total_seconds", namespace="dt")


def test_total_milliseconds(test_expression):
    test_data = [
        timedelta(seconds=1),
        timedelta(milliseconds=1),
        timedelta(microseconds=1),
        timedelta(days=1),
        timedelta(hours=1),
        timedelta(minutes=1),
    ]
    expected = [1000, 1, 0, 86400000, 3600000, 60000]
    test_expression(
        data=test_data, expected=expected, name="total_milliseconds", sql_name="total_milliseconds", namespace="dt"
    )


def test_total_microseconds(test_expression):
    test_data = [
        timedelta(seconds=1),
        timedelta(milliseconds=1),
        timedelta(microseconds=1),
        timedelta(days=1),
        timedelta(hours=1),
        timedelta(minutes=1),
    ]
    expected = [1000000, 1000, 1, 86400000000, 3600000000, 60000000]
    test_expression(
        data=test_data, expected=expected, name="total_microseconds", sql_name="total_microseconds", namespace="dt"
    )


def test_total_nanoseconds(test_expression):
    test_data = [
        timedelta(seconds=1),
        timedelta(milliseconds=1),
        timedelta(microseconds=1),
        timedelta(days=1),
        timedelta(hours=1),
        timedelta(minutes=1),
    ]
    expected = [1000000000, 1000000, 1000, 86400000000000, 3600000000000, 60000000000]
    test_expression(
        data=test_data, expected=expected, name="total_nanoseconds", sql_name="total_nanoseconds", namespace="dt"
    )


def test_total_days(test_expression):
    test_data = [
        timedelta(seconds=1),
        timedelta(milliseconds=1),
        timedelta(microseconds=1),
        timedelta(days=1),
        timedelta(hours=1),
        timedelta(minutes=1),
    ]
    expected = [0, 0, 0, 1, 0, 0]
    test_expression(data=test_data, expected=expected, name="total_days", sql_name="total_days", namespace="dt")


def test_total_hours(test_expression):
    test_data = [
        timedelta(seconds=1),
        timedelta(milliseconds=1),
        timedelta(microseconds=1),
        timedelta(days=1),
        timedelta(hours=1),
        timedelta(minutes=1),
    ]
    expected = [0, 0, 0, 24, 1, 0]
    test_expression(data=test_data, expected=expected, name="total_hours", sql_name="total_hours", namespace="dt")


def test_total_minutes(test_expression):
    test_data = [
        timedelta(seconds=1),
        timedelta(milliseconds=1),
        timedelta(microseconds=1),
        timedelta(days=1),
        timedelta(hours=1),
        timedelta(minutes=1),
    ]
    expected = [0, 0, 0, 1440, 60, 1]
    test_expression(data=test_data, expected=expected, name="total_minutes", sql_name="total_minutes", namespace="dt")



from __future__ import annotations

import itertools
import tempfile
from datetime import date, datetime, timedelta, timezone

import pyarrow as pa
import pytest
import pytz

import daft
from daft import DataType, col

PYARROW_GE_7_0_0 = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) >= (7, 0, 0)


def test_temporal_arithmetic_with_same_type() -> None:
    now = datetime.now()
    now_tz = datetime.now(timezone.utc)
    df = daft.from_pydict(
        {
            "dt_us": [datetime.min, now],
            "dt_us_tz": [datetime.min.replace(tzinfo=timezone.utc), now_tz],
            "date": [datetime.min.date(), now.date()],
            "duration": [timedelta(days=1), timedelta(microseconds=1)],
        }
    )

    df = df.select(
        (df["dt_us"] - df["dt_us"]).alias("zero1"),
        (df["dt_us_tz"] - df["dt_us_tz"]).alias("zero2"),
        (df["date"] - df["date"]).alias("zero3"),
        (df["duration"] + df["duration"]).alias("add_dur"),
        (df["duration"] - df["duration"]).alias("sub_dur"),
    )

    result = df.to_pydict()
    assert result["zero1"] == [timedelta(0), timedelta(0)]
    assert result["zero2"] == [timedelta(0), timedelta(0)]
    assert result["zero3"] == [timedelta(0), timedelta(0)]
    assert result["add_dur"] == [timedelta(days=2), timedelta(microseconds=2)]
    assert result["sub_dur"] == [timedelta(0), timedelta(0)]


@pytest.mark.parametrize("format", ["csv", "parquet"])
def test_temporal_file_roundtrip(format) -> None:
    data = {
        "date32": pa.array([1], pa.date32()),
        "date64": pa.array([1], pa.date64()),
        # Not supported by pyarrow CSV reader yet.
        # "time32_s": pa.array([1], pa.time32("s")),
        # "time32_ms": pa.array([1], pa.time32("ms")),
        # "time64_us": pa.array([1], pa.time64("us")),
        # "time64_ns": pa.array([1], pa.time64("ns")),
        # "duration_s": pa.array([1], pa.duration("s")),
        # "duration_ms": pa.array([1], pa.duration("ms")),
        # "duration_us": pa.array([1], pa.duration("us")),
        # "duration_ns": pa.array([1], pa.duration("ns")),
        # Nanosecond resolution not yet supported (since we currently use Python temporal objects).
        # "timestamp_ns": pa.array([1], pa.timestamp("ns")),
        # "timestamp_ns_tz": pa.array([1], pa.timestamp("ns", tz='UTC')),
        # pyarrow doesn't support writing interval type.
        # "interval": pa.array([pa.scalar((1, 1, 1), type=pa.month_day_nano_interval()).as_py()]),
    }

    # CSV writing of these files only supported by pyarrow CSV writer in PyArrow >= 7.0.0
    if format == "csv" and PYARROW_GE_7_0_0:
        data = {
            **data,
            "timestamp_s": pa.array([1], pa.timestamp("s")),
            "timestamp_ms": pa.array([1], pa.timestamp("ms")),
            "timestamp_us": pa.array([1], pa.timestamp("us")),
            "timestamp_s_utc_tz": pa.array([1], pa.timestamp("s", tz="UTC")),
            "timestamp_ms_utc_tz": pa.array([1], pa.timestamp("ms", tz="UTC")),
            "timestamp_us_utc_tz": pa.array([1], pa.timestamp("us", tz="UTC")),
            "timestamp_s_tz": pa.array([1], pa.timestamp("s", tz="Asia/Singapore")),
            "timestamp_ms_tz": pa.array([1], pa.timestamp("ms", tz="Asia/Singapore")),
            "timestamp_us_tz": pa.array([1], pa.timestamp("us", tz="Asia/Singapore")),
        }

    pa_table = pa.Table.from_pydict(data)

    df = daft.from_arrow(pa_table)

    with tempfile.TemporaryDirectory() as dirname:
        if format == "csv":
            df.write_csv(dirname)
            df_readback = daft.read_csv(dirname).collect()
        elif format == "parquet":
            df.write_parquet(dirname)
            df_readback = daft.read_parquet(dirname).collect()

        assert df.to_pydict() == df_readback.to_pydict()


@pytest.mark.parametrize(
    "timeunit",
    ["s", "ms", "us", "ns"],
)
@pytest.mark.parametrize(
    "timezone",
    [None, "UTC", "America/Los_Angeles", "+04:00"],
)
def test_arrow_timestamp(timeunit, timezone) -> None:
    # Test roundtrip of Arrow timestamps.
    pa_table = pa.Table.from_pydict({"timestamp": pa.array([1, 0, -1], pa.timestamp(timeunit, tz=timezone))})

    df = daft.from_arrow(pa_table)

    assert df.to_arrow() == pa_table


@pytest.mark.skipif(
    not PYARROW_GE_7_0_0,
    reason="PyArrow conversion of timezoned datetime is broken in 6.0.1",
)
@pytest.mark.parametrize("timezone", [None, timezone.utc, timezone(timedelta(hours=-7))])
def test_python_timestamp(timezone) -> None:
    # Test roundtrip of Python timestamps.
    timestamp = datetime.now(timezone)
    df = daft.from_pydict({"timestamp": [timestamp]})

    res = df.to_pydict()["timestamp"][0]
    assert res.isoformat() == timestamp.isoformat()


@pytest.mark.parametrize(
    "timeunit",
    ["s", "ms", "us", "ns"],
)
def test_arrow_duration(timeunit) -> None:
    # Test roundtrip of Arrow timestamps.
    pa_table = pa.Table.from_pydict({"duration": pa.array([1, 0, -1], pa.duration(timeunit))})

    df = daft.from_arrow(pa_table)

    assert df.to_arrow() == pa_table


def test_python_duration() -> None:
    # Test roundtrip of Python durations.
    duration = timedelta(weeks=1, days=1, hours=1, minutes=1, seconds=1, milliseconds=1, microseconds=1)
    df = daft.from_pydict({"duration": [duration]})

    res = df.to_pydict()["duration"][0]
    assert res == duration


def test_temporal_arithmetic_with_duration_lit() -> None:
    df = daft.from_pydict(
        {
            "duration": [timedelta(days=1)],
            "date": [datetime(2021, 1, 1)],
            "timestamp": [datetime(2021, 1, 1)],
        }
    )

    df = df.select(
        (df["date"] + timedelta(days=1)).alias("add_date"),
        (df["date"] - timedelta(days=1)).alias("sub_date"),
        (df["timestamp"] + timedelta(days=1)).alias("add_timestamp"),
        (df["timestamp"] - timedelta(days=1)).alias("sub_timestamp"),
        (df["duration"] + timedelta(days=1)).alias("add_dur"),
        (df["duration"] - timedelta(days=1)).alias("sub_dur"),
    )

    result = df.to_pydict()
    assert result["add_date"] == [datetime(2021, 1, 2)]
    assert result["sub_date"] == [datetime(2020, 12, 31)]
    assert result["add_timestamp"] == [datetime(2021, 1, 2)]
    assert result["sub_timestamp"] == [datetime(2020, 12, 31)]
    assert result["add_dur"] == [timedelta(days=2)]
    assert result["sub_dur"] == [timedelta(0)]


@pytest.mark.parametrize(
    "timeunit",
    ["s", "ms", "us", "ns"],
)
@pytest.mark.parametrize(
    "timezone",
    [None, "UTC"],
)
def test_temporal_arithmetic_timestamp_with_duration(timeunit, timezone) -> None:
    pa_table = pa.Table.from_pydict(
        {
            "timestamp": pa.array([1, 0, -1], pa.timestamp(timeunit, timezone)),
            "duration": pa.array([1, 0, -1], pa.duration(timeunit)),
        }
    )
    df = daft.from_arrow(pa_table)

    df = df.select(
        (df["timestamp"] + df["duration"]).alias("ladd"),
        (df["duration"] + df["timestamp"]).alias("radd"),
        (df["timestamp"] - df["duration"]).alias("sub"),
    ).collect()

    # Check that the result dtypes are correct.
    expected_daft_dtype = daft.DataType.timestamp(daft.TimeUnit.from_str(timeunit), timezone)
    assert df.schema()["ladd"].dtype == expected_daft_dtype
    assert df.schema()["radd"].dtype == expected_daft_dtype
    assert df.schema()["sub"].dtype == expected_daft_dtype

    # Check that the result values are correct.
    expected_result = daft.from_arrow(
        pa.Table.from_pydict(
            {
                "ladd": pa.array([2, 0, -2], pa.timestamp(timeunit, timezone)),
                "radd": pa.array([2, 0, -2], pa.timestamp(timeunit, timezone)),
                "sub": pa.array([0, 0, 0], pa.timestamp(timeunit, timezone)),
            }
        )
    ).to_pydict()

    assert df.to_pydict() == expected_result


def test_temporal_arithmetic_date_with_duration() -> None:
    day_in_seconds = 60 * 60 * 24
    pa_table = pa.Table.from_pydict(
        {
            "date": pa.array([1, 1, 1, 0, -1, -1, -1], pa.date32()),
            "duration": pa.array(
                [
                    day_in_seconds,
                    day_in_seconds - 1,
                    day_in_seconds + 1,
                    0,
                    -day_in_seconds,
                    -day_in_seconds + 1,
                    -day_in_seconds - 1,
                ],
                pa.duration("s"),
            ),
        }
    )
    df = daft.from_arrow(pa_table)

    df = df.select(
        (df["date"] + df["duration"]).alias("ladd"),
        (df["duration"] + df["date"]).alias("radd"),
        (df["date"] - df["duration"]).alias("sub"),
    ).collect()

    # Check that the result dtypes are correct.
    expected_daft_dtype = daft.DataType.date()
    assert df.schema()["ladd"].dtype == expected_daft_dtype
    assert df.schema()["radd"].dtype == expected_daft_dtype
    assert df.schema()["sub"].dtype == expected_daft_dtype

    # Check that the result values are correct.
    expected_result = daft.from_arrow(
        pa.Table.from_pydict(
            {
                "ladd": pa.array([2, 1, 2, 0, -2, -1, -2], pa.date32()),
                "radd": pa.array([2, 1, 2, 0, -2, -1, -2], pa.date32()),
                "sub": pa.array([0, 1, 0, 0, 0, -1, 0], pa.date32()),
            }
        )
    ).to_pydict()

    assert df.to_pydict() == expected_result


@pytest.mark.parametrize(
    "t_timeunit",
    ["s", "ms", "us", "ns"],
)
@pytest.mark.parametrize(
    "d_timeunit",
    ["s", "ms", "us", "ns"],
)
@pytest.mark.parametrize(
    "timezone",
    [None, "UTC"],
)
def test_temporal_arithmetic_mismatch_granularity(t_timeunit, d_timeunit, timezone) -> None:
    if t_timeunit == d_timeunit:
        return

    pa_table = pa.Table.from_pydict(
        {
            "timestamp": pa.array([1, 0, -1], pa.timestamp(t_timeunit, timezone)),
            "duration": pa.array([1, 0, -1], pa.duration(d_timeunit)),
        }
    )

    df = daft.from_arrow(pa_table)
    for expression in [
        (df["timestamp"] + df["duration"]).alias("ladd"),
        (df["duration"] + df["timestamp"]).alias("radd"),
        (df["timestamp"] - df["duration"]).alias("sub"),
    ]:
        with pytest.raises(ValueError):
            df.select(expression).collect()


@pytest.mark.parametrize("tu1, tu2", itertools.product(["ns", "us", "ms"], repeat=2))
@pytest.mark.parametrize("tz_repr", ["UTC", "+00:00"])
def test_join_timestamp_same_timezone(tu1, tu2, tz_repr):
    tz1 = [datetime(2022, 1, 1, tzinfo=pytz.utc), datetime(2022, 2, 1, tzinfo=pytz.utc)]
    tz2 = [datetime(2022, 1, 2, tzinfo=pytz.utc), datetime(2022, 1, 1, tzinfo=pytz.utc)]
    df1 = daft.from_pydict({"t": tz1, "x": [1, 2]}).with_column("t", col("t").cast(DataType.timestamp(tu1, tz_repr)))
    df2 = daft.from_pydict({"t": tz2, "y": [3, 4]}).with_column("t", col("t").cast(DataType.timestamp(tu2, tz_repr)))
    res = df1.join(df2, on="t")
    assert res.to_pydict() == {
        "t": [datetime(2022, 1, 1, tzinfo=pytz.utc)],
        "x": [1],
        "y": [4],
    }


@pytest.mark.parametrize(
    "op,expected",
    [
        (
            (col("datetimes") + daft.interval(years=1)),
            [
                datetime(2022, 1, 1, 0, 0),
                datetime(2022, 1, 2, 0, 0),
                datetime(2021, 3, 1, 0, 0),
                datetime(2021, 2, 28, 0, 0),
            ],
        ),
        (
            (col("datetimes") + daft.interval(months=1)),
            [
                datetime(2021, 2, 1, 0, 0),
                datetime(2021, 2, 2, 0, 0),
                datetime(2020, 3, 29, 0, 0),
                datetime(2020, 3, 28, 0, 0),
            ],
        ),
        (
            (col("datetimes") + daft.interval(days=1)),
            [
                datetime(2021, 1, 2, 0, 0),
                datetime(2021, 1, 3, 0, 0),
                datetime(2020, 3, 1, 0, 0),
                datetime(2020, 2, 29, 0, 0),
            ],
        ),
        (
            (col("datetimes") + daft.interval(hours=1)),
            [
                datetime(2021, 1, 1, 1, 0),
                datetime(2021, 1, 2, 1, 0),
                datetime(2020, 2, 29, 1, 0),
                datetime(2020, 2, 28, 1, 0),
            ],
        ),
        (
            (col("datetimes") + daft.interval(minutes=1)),
            [
                datetime(2021, 1, 1, 0, 1),
                datetime(2021, 1, 2, 0, 1),
                datetime(2020, 2, 29, 0, 1),
                datetime(2020, 2, 28, 0, 1),
            ],
        ),
        (
            (col("datetimes") + daft.interval(seconds=1)),
            [
                datetime(2021, 1, 1, 0, 0, 1),
                datetime(2021, 1, 2, 0, 0, 1),
                datetime(2020, 2, 29, 0, 0, 1),
                datetime(2020, 2, 28, 0, 0, 1),
            ],
        ),
        (
            (col("datetimes") + daft.interval(millis=1)),
            [
                datetime(2021, 1, 1, 0, 0, 0, 1000),
                datetime(2021, 1, 2, 0, 0, 0, 1000),
                datetime(2020, 2, 29, 0, 0, 0, 1000),
                datetime(2020, 2, 28, 0, 0, 0, 1000),
            ],
        ),
        (
            (col("datetimes") - daft.interval(years=1)),
            [
                datetime(2020, 1, 2, 0, 0),
                datetime(2020, 1, 3, 0, 0),
                datetime(2019, 2, 28, 0, 0),
                datetime(2019, 2, 27, 0, 0),
            ],
        ),
        (
            (col("datetimes") - daft.interval(months=1)),
            [
                datetime(2020, 12, 1, 0, 0),
                datetime(2020, 12, 2, 0, 0),
                datetime(2020, 1, 31, 0, 0),
                datetime(2020, 1, 30, 0, 0),
            ],
        ),
        (
            (col("datetimes") - daft.interval(days=1)),
            [
                datetime(2021, 1, 2, 0, 0),
                datetime(2021, 1, 3, 0, 0),
                datetime(2020, 3, 1, 0, 0),
                datetime(2020, 2, 29, 0, 0),
            ],
        ),
        (
            (col("datetimes") - daft.interval(hours=1)),
            [
                datetime(2020, 12, 31, 23, 0),
                datetime(2021, 1, 1, 23, 0),
                datetime(2020, 2, 28, 23, 0),
                datetime(2020, 2, 27, 23, 0),
            ],
        ),
        (
            col("datetimes") - daft.interval(minutes=1),
            [
                datetime(2020, 12, 31, 23, 59),
                datetime(2021, 1, 1, 23, 59),
                datetime(2020, 2, 28, 23, 59),
                datetime(2020, 2, 27, 23, 59),
            ],
        ),
        (
            (col("datetimes") - daft.interval(seconds=1)),
            [
                datetime(2020, 12, 31, 23, 59, 59),
                datetime(2021, 1, 1, 23, 59, 59),
                datetime(2020, 2, 28, 23, 59, 59),
                datetime(2020, 2, 27, 23, 59, 59),
            ],
        ),
        (
            (col("datetimes") - daft.interval(millis=1)),
            [
                datetime(2020, 12, 31, 23, 59, 59, 999000),
                datetime(2021, 1, 1, 23, 59, 59, 999000),
                datetime(2020, 2, 28, 23, 59, 59, 999000),
                datetime(2020, 2, 27, 23, 59, 59, 999000),
            ],
        ),
    ],
)
def test_intervals(op, expected):
    datetimes = [
        datetime(2021, 1, 1, 0, 0, 0),
        datetime(2021, 1, 2, 0, 0, 0),
        # add a datetime with a leap event
        datetime(2020, 2, 29, 0, 0, 0),
        # and another one that should land on a leap event
        datetime(2020, 2, 28, 0, 0, 0),
    ]
    actual = (
        daft.from_pydict(
            {
                "datetimes": datetimes,
            }
        )
        .select(op)
        .collect()
        .to_pydict()
    )
    expected = {"datetimes": expected}

    assert actual == expected


@pytest.mark.parametrize(
    "value",
    [
        date(2020, 1, 1),  # explicit date
        "2020-01-01",  # implicit coercion
    ],
)
def test_date_comparison(value):
    date_df = daft.from_pydict({"date_str": ["2020-01-01", "2020-01-02", "2020-01-03"]})
    date_df = date_df.with_column("date", col("date_str").str.to_date("%Y-%m-%d"))
    actual = date_df.filter(col("date") == value).select("date").to_pydict()

    expected = {"date": [date(2020, 1, 1)]}

    assert actual == expected


def test_date_and_datetime_day_of_week():
    df = daft.from_pydict(
        {
            "date": [date(2020, 1, 1), date(2020, 12, 31), date(2021, 12, 31)],
            "datetime": [
                datetime(2020, 1, 1, 0, 0, 0),
                datetime(2020, 12, 31, 23, 59, 59),
                datetime(2021, 12, 31, 23, 59, 59),
            ],
        }
    )

    df = df.select(df["date"].dt.day_of_week().alias("date_dow"), df["datetime"].dt.day_of_week().alias("datetime_dow"))

    expected = {"date_dow": [2, 3, 4], "datetime_dow": [2, 3, 4]}

    assert df.to_pydict() == expected


def test_date_and_datetime_day_of_month():
    df = daft.from_pydict(
        {
            "date": [date(2020, 1, 1), date(2020, 12, 31), date(2021, 12, 31)],
            "datetime": [
                datetime(2020, 1, 1, 0, 0, 0),
                datetime(2020, 12, 31, 23, 59, 59),
                datetime(2021, 12, 31, 23, 59, 59),
            ],
        }
    )

    df = df.select(
        df["date"].dt.day_of_month().alias("date_dom"), df["datetime"].dt.day_of_month().alias("datetime_dom")
    )

    expected = {"date_dom": [1, 31, 31], "datetime_dom": [1, 31, 31]}

    assert df.to_pydict() == expected


def test_date_and_datetime_day_of_year():
    df = daft.from_pydict(
        {
            "date": [date(2020, 1, 1), date(2020, 12, 31), date(2021, 12, 31)],
            "datetime": [
                datetime(2020, 1, 1, 0, 0, 0),
                datetime(2020, 12, 31, 23, 59, 59),
                datetime(2021, 12, 31, 23, 59, 59),
            ],
        }
    )

    df = df.select(df["date"].dt.day_of_year().alias("date_doy"), df["datetime"].dt.day_of_year().alias("datetime_doy"))

    expected = {"date_doy": [1, 366, 365], "datetime_doy": [1, 366, 365]}

    assert df.to_pydict() == expected


def test_date_and_datetime_week_of_year():
    df = daft.from_pydict(
        {
            "date": [date(2020, 1, 1), date(2020, 12, 31), date(2021, 12, 31)],
            "datetime": [
                datetime(2020, 1, 1, 0, 0, 0),
                datetime(2020, 12, 31, 23, 59, 59),
                datetime(2021, 12, 31, 23, 59, 59),
            ],
        }
    )

    df = df.select(
        df["date"].dt.week_of_year().alias("date_woy"), df["datetime"].dt.week_of_year().alias("datetime_woy")
    )

    expected = {"date_woy": [1, 53, 52], "datetime_woy": [1, 53, 52]}

    assert df.to_pydict() == expected


def test_date_to_unix_epoch():
    df = daft.from_pydict(
        {
            "date": [date(2020, 1, 1), date(2020, 12, 31), date(2021, 12, 31)],
            "datetime": [
                datetime(2020, 1, 1, 0, 0, 0),
                datetime(2020, 12, 31, 23, 59, 59),
                datetime(2021, 12, 31, 23, 59, 59),
            ],
        }
    )
    actual = df.select(
        df["date"].dt.to_unix_epoch().alias("date_epoch"),
        df["date"].dt.to_unix_epoch("s").alias("date_epoch_s"),
        df["date"].dt.to_unix_epoch("ms").alias("date_epoch_ms"),
        df["date"].dt.to_unix_epoch("us").alias("date_epoch_us"),
        df["date"].dt.to_unix_epoch("ns").alias("date_epoch_ns"),
        df["datetime"].dt.to_unix_epoch().alias("datetime_epoch"),
        df["datetime"].dt.to_unix_epoch("s").alias("datetime_epoch_s"),
        df["datetime"].dt.to_unix_epoch("ms").alias("datetime_epoch_ms"),
        df["datetime"].dt.to_unix_epoch("us").alias("datetime_epoch_us"),
        df["datetime"].dt.to_unix_epoch("ns").alias("datetime_epoch_ns"),
    ).to_pydict()

    expected = {
        "date_epoch": [1577836800, 1609372800, 1640908800],
        "date_epoch_s": [1577836800, 1609372800, 1640908800],
        "date_epoch_ms": [1577836800000, 1609372800000, 1640908800000],
        "date_epoch_us": [1577836800000000, 1609372800000000, 1640908800000000],
        "date_epoch_ns": [1577836800000000000, 1609372800000000000, 1640908800000000000],
        "datetime_epoch": [1577836800, 1609459199, 1640995199],
        "datetime_epoch_s": [1577836800, 1609459199, 1640995199],
        "datetime_epoch_ms": [1577836800000, 1609459199000, 1640995199000],
        "datetime_epoch_us": [1577836800000000, 1609459199000000, 1640995199000000],
        "datetime_epoch_ns": [1577836800000000000, 1609459199000000000, 1640995199000000000],
    }

    assert actual == expected


def test_date_to_string():
    df = daft.from_pydict(
        {
            "dates": [date(2023, 1, 1), date(2023, 1, 2), date(2023, 1, 3)],
            "datetimes": [
                datetime(2023, 1, 1, 12, 1),
                datetime(2023, 1, 2, 12, 0, 0, 0),
                datetime(2023, 1, 3, 12, 0, 0, 999_999),
            ],
        }
    )

    df = df.with_column("datetimes_s", daft.col("datetimes").cast(daft.DataType.timestamp("s")))

    actual = df.select(
        daft.col("dates").dt.strftime().alias("iso_date"),
        daft.col("dates").dt.strftime(format="%m/%d/%Y").alias("custom_date"),
        daft.col("datetimes_s").dt.strftime().alias("iso_datetime"),
        daft.col("datetimes_s").dt.strftime(format="%Y/%m/%d %H:%M:%S").alias("custom_datetime"),
    ).to_pydict()

    expected = {
        "iso_date": ["2023-01-01", "2023-01-02", "2023-01-03"],
        "custom_date": ["01/01/2023", "01/02/2023", "01/03/2023"],
        "iso_datetime": ["2023-01-01T12:01:00", "2023-01-02T12:00:00", "2023-01-03T12:00:00"],
        "custom_datetime": ["2023/01/01 12:01:00", "2023/01/02 12:00:00", "2023/01/03 12:00:00"],
    }

    assert actual == expected


@pytest.mark.parametrize(
    "timeunit",
    ["s", "ms", "us", "ns", "seconds", "milliseconds", "microseconds", "nanoseconds"],
)
def test_date_to_unix_epoch_valid_timeunits(timeunit):
    df = daft.from_pydict(
        {
            "date": [date(2020, 1, 1), date(2020, 12, 31), date(2021, 12, 31)],
        }
    )
    try:
        df.select(
            df["date"].dt.to_unix_epoch(timeunit).alias("date_epoch"),
        ).to_pydict()
    except ValueError:
        pytest.fail(f"to_unix_epoch with timeunit {timeunit} raised an exception.")


@pytest.mark.parametrize(
    "timeunit",
    ["second", "millis", "nanos", "millisecond", "nanosecond", "millis", "micros"],
)
def test_date_to_unix_epoch_invalid_timeunits(timeunit):
    df = daft.from_pydict(
        {
            "date": [date(2020, 1, 1)],
        }
    )
    try:
        df.select(
            df["date"].dt.to_unix_epoch(timeunit).alias("date_epoch"),
        ).to_pydict()
        pytest.fail(f"to_unix_epoch with timeunit {timeunit} did not raise an exception.")
    except ValueError:
        pass


@pytest.mark.parametrize(
    "fmt,expected",
    [
        ("%Y-%m-%d", ["2023-01-01", "2023-01-02", "2023-01-03"]),
        ("%m/%d/%Y", ["01/01/2023", "01/02/2023", "01/03/2023"]),
        ("%Y/%m/%d %H:%M:%S", ["2023/01/01 12:01:00", "2023/01/02 12:00:00", "2023/01/03 12:00:00"]),
        ("%c", ["Sun Jan  1 12:01:00 2023", "Mon Jan  2 12:00:00 2023", "Tue Jan  3 12:00:00 2023"]),
        ("%x", ["01/01/23", "01/02/23", "01/03/23"]),
    ],
)
def test_datetime_to_string(fmt, expected):
    df = daft.from_pydict(
        {
            "dates": [date(2023, 1, 1), date(2023, 1, 2), date(2023, 1, 3)],
            "datetimes": [
                datetime(2023, 1, 1, 12, 1),
                datetime(2023, 1, 2, 12, 0, 0, 0),
                datetime(2023, 1, 3, 12, 0, 0, 999_999),
            ],
        }
    )
    df = df.with_column("datetimes_s", daft.col("datetimes").cast(daft.DataType.timestamp("s")))

    actual = df.select(
        daft.col("datetimes").dt.strftime(fmt).alias("formatted_datetime"),
    ).to_pydict()

    expected = {
        "formatted_datetime": expected,
    }

    assert actual == expected


@pytest.mark.parametrize(
    "fmt,expected",
    [
        ("%H-%M-%S", ["01-02-03", "02-03-04", "03-04-05"]),
        ("%H:%M:%S", ["01:02:03", "02:03:04", "03:04:05"]),
        ("H:%H, M:%M, S:%S", ["H:01, M:02, S:03", "H:02, M:03, S:04", "H:03, M:04, S:05"]),
    ],
)
def test_time_to_string(fmt, expected):
    from datetime import time

    df = daft.from_pydict(
        {
            "times": [time(1, 2, 3), time(2, 3, 4), time(3, 4, 5)],
        }
    )

    actual = df.select(
        daft.col("times").dt.strftime(fmt).alias("formatted_time"),
    ).to_pydict()

    expected = {
        "formatted_time": expected,
    }

    assert actual == expected


@pytest.mark.parametrize(
    "value",
    [
        1,
        True,
        "foo",
        1.0,
        [1],
        [True],
        [1.0],
    ],
)
def test_datetime_to_string_errors(value):
    df = daft.from_pydict({"invalid": [value]})

    with pytest.raises(daft.exceptions.DaftCoreException):
        df.select(daft.col("invalid").dt.strftime(format="%Y-%m-%d")).to_pydict()



from __future__ import annotations

import datetime
from decimal import Decimal

import pytest

from daft import col
from daft.recordbatch import MicroPartition


@pytest.mark.parametrize(
    "value,lower,upper,expected",
    [
        pytest.param([1, 2, 3, 4], 1, 2, [True, True, False, False], id="IntIntInt"),
        pytest.param([1, 2, 3, 4], 1.0, 2.0, [True, True, False, False], id="IntFloatFloat"),
        pytest.param([1, 2, 3, 4], 1, 2.0, [True, True, False, False], id="IntIntFloat"),
        pytest.param(
            [Decimal("1.0"), Decimal("2.0"), Decimal("3.0"), Decimal("4.0")],
            Decimal("1.0"),
            Decimal("2.0"),
            [True, True, False, False],
            id="DecimalDecimalDecimal",
        ),
        pytest.param(
            [Decimal("1.0"), Decimal("2.0"), Decimal("3.0"), Decimal("4.0")],
            1,
            Decimal("2.0"),
            [True, True, False, False],
            id="DecimalIntDecimal",
        ),
        pytest.param(
            [Decimal("1.0"), Decimal("2.0"), Decimal("3.0"), Decimal("4.0")],
            1.0,
            2.0,
            [True, True, False, False],
            id="DecimalFloatFloat",
        ),
        pytest.param([1.0, 2.0, 3.0, 4.0], 1.0, 2.0, [True, True, False, False], id="FloatFloatFloat"),
        pytest.param([1.0, 2.0, 3.0, 4.0], 1, 2, [True, True, False, False], id="FloatIntInt"),
        pytest.param([1.0, 2.0, 3.0, 4.0], 1, 2.0, [True, True, False, False], id="FloatIntFloat"),
        pytest.param([1.0, 2.0, 3.0, 4.0], None, 1, [None, False, False, False], id="FloatNullInt"),
        pytest.param([1.0, 2.0, 3.0, 4.0], 1, None, [None, None, None, None], id="FloatIntNull"),
        pytest.param([1.0, 2.0, 3.0, 4.0], None, None, [None, None, None, None], id="FloatNullNull"),
        pytest.param([None, None, None, None], None, None, [None, None, None, None], id="NullNullNull"),
        pytest.param([None, None, None, None], 1, 1, [None, None, None, None], id="NullIntInt"),
        pytest.param(
            [datetime.datetime(2023, 1, 1), datetime.datetime(2022, 1, 1)],
            datetime.datetime(2022, 12, 30),
            datetime.datetime(2023, 1, 2),
            [True, False],
            id="Datetime",
        ),
    ],
)
def test_table_expr_between_scalars(value, lower, upper, expected) -> None:
    daft_recordbatch = MicroPartition.from_pydict({"value": value})
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("value").between(lower, upper)])
    pydict = daft_recordbatch.to_pydict()
    assert pydict["value"] == expected


@pytest.mark.parametrize(
    "value,lower,upper,expected",
    [
        pytest.param([1, 2, 3, 4], [1, 1, 1, 1], [2, 2, 2, 2], [True, True, False, False], id="IntIntInt"),
        pytest.param(
            [1, 2, 3, 4], [1.0, 1.0, 1.0, 1.0], [2.0, 2.0, 2.0, 2.0], [True, True, False, False], id="IntFloatFloat"
        ),
        pytest.param(
            [Decimal("1.0"), Decimal("2.0"), Decimal("3.0"), Decimal("4.0")],
            [Decimal("1.0"), Decimal("1.0"), Decimal("1.0"), Decimal("1.0")],
            [Decimal("2.0"), Decimal("2.0"), Decimal("2.0"), Decimal("2.0")],
            [True, True, False, False],
            id="DecimalDecimalDecimal",
        ),
        pytest.param(
            [Decimal("1.0"), Decimal("2.0"), Decimal("3.0"), Decimal("4.0")],
            [1, 1, 1, 1],
            [2.0, 2.0, 2.0, 2.0],
            [True, True, False, False],
            id="DecimalIntFloat",
        ),
        pytest.param([1, 2, 3, 4], [1, 1, 1, 1], [2.0, 2.0, 2.0, 2.0], [True, True, False, False], id="IntIntFloat"),
        pytest.param(
            [None, None, None, None],
            [2.0, 2.0, 2.0, 2.0],
            [2.0, 2.0, 2.0, 2.0],
            [None, None, None, None],
            id="NullFloatFloat",
        ),
        pytest.param(
            [None, None, None, None],
            [None, None, None, None],
            [2.0, 2.0, 2.0, 2.0],
            [None, None, None, None],
            id="NullNullFloat",
        ),
        pytest.param(
            [None, None, None, None],
            [2.0, 2.0, 2.0, 2.0],
            [None, None, None, None],
            [None, None, None, None],
            id="NullFloatNull",
        ),
        pytest.param(
            [None, None, None, None],
            [None, None, None, None],
            [None, None, None, None],
            [None, None, None, None],
            id="NullNullNull",
        ),
        pytest.param(
            [1.0, 2.0, 3.0, 4.0],
            [1.0, 1.0, 1.0, 1.0],
            [2.0, 2.0, 2.0, 2.0],
            [True, True, False, False],
            id="FloatFloatFloat",
        ),
        pytest.param([1.0, 2.0, 3.0, 4.0], [1, 1, 1, 1], [2, 2, 2, 2], [True, True, False, False], id="FloatIntInt"),
        pytest.param(
            [1.0, 2.0, 3.0, 4.0], [1, 1, 1, 1], [2.0, 2.0, 2.0, 2.0], [True, True, False, False], id="FloatIntFloat"
        ),
        pytest.param(
            [datetime.datetime(2023, 1, 1), datetime.datetime(2022, 1, 1)],
            [datetime.datetime(2022, 12, 30), datetime.datetime(2022, 12, 30)],
            [datetime.datetime(2023, 1, 2), datetime.datetime(2023, 1, 2)],
            [True, False],
            id="Datetime",
        ),
    ],
)
def test_between_columns(value, lower, upper, expected) -> None:
    table = {"value": value, "lower": lower, "upper": upper}
    daft_recordbatch = MicroPartition.from_pydict(table)
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("value").between(col("lower"), col("upper"))])
    pydict = daft_recordbatch.to_pydict()
    assert pydict["value"] == expected


@pytest.mark.parametrize(
    "value,lower,upper",
    [
        pytest.param(["str1", "str2"], 1, 2, id="StrIntInt"),
        pytest.param([1, 2], "str", 1, id="IntStrInt"),
    ],
)
def test_between_between_different_types(value, lower, upper) -> None:
    daft_recordbatch = MicroPartition.from_pydict({"a": value})
    with pytest.raises(ValueError):
        daft_recordbatch = daft_recordbatch.eval_expression_list([col("a").between(lower, upper)])


def test_between_bad_input() -> None:
    daft_recordbatch = MicroPartition.from_pydict({"a": [1, 2, 3]})
    with pytest.raises(TypeError):
        daft_recordbatch = daft_recordbatch.eval_expression_list([col("a").between([1, 2, 3], 1)])


@pytest.mark.parametrize(
    "value,lower,upper,expected",
    [
        pytest.param([1, 2, 3, 4], [1, 2, 3, 4], 2, [True, True, False, False], id="IntIntInt"),
        pytest.param([1, 2, 3, 4], [1.0, 2.0, 3.0, 4.0], 2.0, [True, True, False, False], id="IntFloatFloat"),
        pytest.param([1, 2, 3, 4], [1, 2, 3, 4], 2.0, [True, True, False, False], id="IntIntFloat"),
        pytest.param([1.0, 2.0, 3.0, 4.0], [1.0, 2.0, 3.0, 4.0], 2.0, [True, True, False, False], id="FloatFloatFloat"),
        pytest.param([1.0, 2.0, 3.0, 4.0], [1.0, 2.0, 3.0, 4.0], 2, [True, True, False, False], id="FloatIntInt"),
        pytest.param([1.0, 2.0, 3.0, 4.0], [1, 2, 3, 4], 2.0, [True, True, False, False], id="FloatIntFloat"),
        pytest.param([1.0, 2.0, 3.0, 4.0], [None, None, None, None], 1, [None, False, False, False], id="FloatNullInt"),
        pytest.param([1.0, 2.0, 3.0, 4.0], [1, 2, 3, 4], None, [None, None, None, None], id="FloatIntNull"),
        pytest.param(
            [1.0, 2.0, 3.0, 4.0], [None, None, None, None], None, [None, None, None, None], id="FloatNullNull"
        ),
        pytest.param(
            [None, None, None, None], [None, None, None, None], None, [None, None, None, None], id="NullNullNull"
        ),
        pytest.param([None, None, None, None], [1, 2, 3, 4], 1, [None, None, None, None], id="NullIntInt"),
        pytest.param(
            [datetime.datetime(2023, 1, 1), datetime.datetime(2022, 1, 1)],
            [datetime.datetime(2022, 12, 30), datetime.datetime(2022, 12, 30)],
            datetime.datetime(2023, 1, 2),
            [True, False],
            id="Datetime",
        ),
    ],
)
def test_table_expr_between_col_and_scalar(value, lower, upper, expected) -> None:
    table = {"value": value, "lower": lower}
    daft_recordbatch = MicroPartition.from_pydict(table)
    daft_recordbatch = daft_recordbatch.eval_expression_list([col("value").between(col("lower"), upper)])
    pydict = daft_recordbatch.to_pydict()
    assert pydict["value"] == expected



from __future__ import annotations

import numpy as np
import pytest

import daft
from daft import DataFrame

NUM_SAMPLES = 10_000_000
NUM_PARTITIONS = 100


@pytest.mark.benchmark(group="partitioning")
@pytest.mark.parametrize("end_partitions", [NUM_PARTITIONS])
@pytest.mark.parametrize("num_samples", [NUM_SAMPLES])
def test_split(benchmark, num_samples, end_partitions) -> None:
    """Test performance of splitting into multiple partitions."""
    df = daft.from_pydict({"mycol": np.arange(num_samples)}).collect()

    # Run the benchmark.
    def bench() -> DataFrame:
        return df.into_partitions(end_partitions).collect()

    benchmark(bench)


@pytest.mark.benchmark(group="partitioning")
@pytest.mark.parametrize("start_partitions", [NUM_PARTITIONS])
@pytest.mark.parametrize("num_samples", [NUM_SAMPLES])
def test_coalesce(benchmark, num_samples, start_partitions) -> None:
    """Test performance of coalescing partitions."""
    df = daft.from_pydict({"mycol": np.arange(num_samples)}).into_partitions(start_partitions).collect()

    # Run the benchmark.
    def bench() -> DataFrame:
        return df.into_partitions(1).collect()

    benchmark(bench)


@pytest.mark.benchmark(group="partitioning")
@pytest.mark.parametrize("end_partitions", [NUM_PARTITIONS])
@pytest.mark.parametrize("start_partitions", [NUM_PARTITIONS])
@pytest.mark.parametrize("num_samples", [NUM_SAMPLES])
def test_repartition_random(benchmark, num_samples, start_partitions, end_partitions) -> None:
    """Test performance of random repartitioning."""
    df = daft.from_pydict({"mycol": np.arange(num_samples)}).into_partitions(start_partitions).collect()

    # Run the benchmark.
    def bench() -> DataFrame:
        return df.repartition(end_partitions).collect()

    benchmark(bench)


@pytest.mark.benchmark(group="partitioning")
@pytest.mark.parametrize("distribution", ["uniform", "skewed"])
@pytest.mark.parametrize("end_partitions", [NUM_PARTITIONS])
@pytest.mark.parametrize("start_partitions", [1, NUM_PARTITIONS])
@pytest.mark.parametrize("num_samples", [NUM_SAMPLES])
def test_repartition_hash(benchmark, num_samples, start_partitions, end_partitions, distribution) -> None:
    """Test performance of hash repartitioning."""
    data = np.arange(num_samples)
    if distribution == "skewed":
        data[: num_samples // 2] = 0

    np.random.shuffle(data)

    df = daft.from_pydict({"mycol": data}).into_partitions(start_partitions).collect()

    # Run the benchmark.
    def bench() -> DataFrame:
        return df.repartition(end_partitions, "mycol").collect()

    benchmark(bench)



from __future__ import annotations

import copy
import io

import cv2
import numpy as np
import pyarrow as pa
import pytest
from PIL import Image, ImageSequence

from daft.datatype import DataType, get_super_ext_type
from daft.series import Series

DaftExtension = get_super_ext_type()

MODE_TO_NP_DTYPE = {
    "L": np.uint8,
    "LA": np.uint8,
    "RGB": np.uint8,
    "RGBA": np.uint8,
    "L16": np.uint16,
    "LA16": np.uint16,
    "RGB16": np.uint16,
    "RGBA16": np.uint16,
    "RGB32F": np.float32,
    "RGBA32F": np.float32,
}

MODE_TO_NUM_CHANNELS = {
    "L": 1,
    "LA": 2,
    "RGB": 3,
    "RGBA": 4,
    "L16": 1,
    "LA16": 2,
    "RGB16": 3,
    "RGBA16": 4,
    "RGB32F": 3,
    "RGBA32F": 4,
}

NUM_CHANNELS_TO_MODE = {
    1: "L",
    2: "LA",
    3: "RGB",
    4: "RGBA",
}

MODE_TO_OPENCV_COLOR_CONVERSION = {
    "RGB": cv2.COLOR_RGB2BGR,
    "RGBA": cv2.COLOR_RGBA2BGRA,
    "RGB16": cv2.COLOR_RGB2BGR,
    "RGBA16": cv2.COLOR_RGBA2BGRA,
    "RGB32F": cv2.COLOR_RGB2BGR,
    "RGBA32F": cv2.COLOR_RGBA2BGRA,
}

MODE_TO_OPENCV_COLOR_CONVERSION_DECODE = {
    "RGB": cv2.COLOR_BGR2RGB,
    "RGBA": cv2.COLOR_BGRA2RGBA,
    "RGB16": cv2.COLOR_BGR2RGB,
    "RGBA16": cv2.COLOR_BGRA2RGBA,
    "RGB32F": cv2.COLOR_BGR2RGB,
    "RGBA32F": cv2.COLOR_BGRA2RGBA,
}


@pytest.mark.parametrize("give_mode", [True, False])
def test_image_round_trip(give_mode):
    data = [
        np.arange(12, dtype=np.uint8).reshape((2, 2, 3)),
        np.arange(12, 39, dtype=np.uint8).reshape((3, 3, 3)),
        None,
    ]
    s = Series.from_pylist(data, pyobj="force")

    mode = "RGB" if give_mode else None
    target_dtype = DataType.image(mode)

    t = s.cast(target_dtype)

    assert t.datatype() == target_dtype

    # Test pylist roundtrip.
    back_dtype = DataType.python()
    back = t.cast(back_dtype)

    assert back.datatype() == back_dtype

    out = back.to_pylist()
    np.testing.assert_equal(out, data)

    # Test Arrow roundtrip.
    arrow_arr = t.to_arrow()

    assert isinstance(arrow_arr.type, DaftExtension)
    from_arrow = Series.from_arrow(t.to_arrow())

    assert from_arrow.datatype() == t.datatype()
    np.testing.assert_equal(from_arrow.to_pylist(), t.to_pylist())

    t_copy = copy.deepcopy(t)
    assert t_copy.datatype() == t.datatype()
    np.testing.assert_equal(t_copy.to_pylist(), t.to_pylist())


def test_fixed_shape_image_round_trip():
    height = 2
    width = 2
    shape = (height, width, 3)
    data = [
        np.arange(12, dtype=np.uint8).reshape(shape),
        np.arange(12, 24, dtype=np.uint8).reshape(shape),
        None,
    ]
    s = Series.from_pylist(data, pyobj="force")

    target_dtype = DataType.image("RGB", height, width)

    t = s.cast(target_dtype)

    assert t.datatype() == target_dtype

    # Test pylist roundtrip.
    back_dtype = DataType.python()
    back = t.cast(back_dtype)

    assert back.datatype() == back_dtype

    out = back.to_pylist()
    np.testing.assert_equal(out, data)

    # Test Arrow roundtrip.
    arrow_arr = t.to_arrow()

    assert isinstance(arrow_arr.type, DaftExtension)
    from_arrow = Series.from_arrow(t.to_arrow())

    assert from_arrow.datatype() == t.datatype()
    np.testing.assert_equal(from_arrow.to_pylist(), t.to_pylist())

    t_copy = copy.deepcopy(t)
    assert t_copy.datatype() == t.datatype()
    np.testing.assert_equal(t_copy.to_pylist(), t.to_pylist())


@pytest.mark.parametrize(
    "mode",
    [
        "L",
        "LA",
        "RGB",
        "RGBA",
    ],
)
@pytest.mark.parametrize("fixed_shape", [True, False])
def test_image_pil_inference(fixed_shape, mode):
    np_dtype = MODE_TO_NP_DTYPE[mode]
    num_channels = MODE_TO_NUM_CHANNELS[mode]
    if fixed_shape:
        height = 4
        width = 4
        shape = (height, width)
        if num_channels > 1:
            shape += (num_channels,)
        arr = np.arange(np.prod(shape)).reshape(shape).astype(np_dtype)
        arrs = [arr, arr, None]
    else:
        shape1 = (2, 2)
        shape2 = (3, 3)
        if num_channels > 1:
            shape1 += (num_channels,)
            shape2 += (num_channels,)
        arr1 = np.arange(np.prod(shape1)).reshape(shape1).astype(np_dtype)
        arr2 = np.arange(np.prod(shape1), np.prod(shape1) + np.prod(shape2)).reshape(shape2).astype(np_dtype)
        arrs = [arr1, arr2, None]
    if mode in ("LA", "RGBA"):
        for arr in arrs:
            if arr is not None:
                arr[..., -1] = 255
    imgs = [Image.fromarray(arr, mode=mode) if arr is not None else None for arr in arrs]
    s = Series.from_pylist(imgs, pyobj="allow")
    assert s.datatype() == DataType.image(mode)
    out = s.to_pylist()
    if num_channels == 1:
        arrs = [np.expand_dims(arr, -1) for arr in arrs]
    np.testing.assert_equal(out, arrs)


def test_image_pil_inference_mixed():
    rgba = np.ones((2, 2, 4), dtype=np.uint8)
    rgba[..., 1] = 2
    rgba[..., 2] = 3
    rgba[..., 3] = 4

    arrs = [
        rgba[..., :3],  # RGB
        rgba,  # RGBA
        np.arange(12, dtype=np.uint8).reshape((1, 4, 3)),  # RGB
        np.arange(12, dtype=np.uint8).reshape((3, 4)) * 10,  # L
        np.ones(24, dtype=np.uint8).reshape((3, 4, 2)) * 10,  # LA
        None,
    ]
    imgs = [
        Image.fromarray(arr, mode=NUM_CHANNELS_TO_MODE[arr.shape[-1] if arr.ndim == 3 else 1])
        if arr is not None
        else None
        for arr in arrs
    ]

    # Forcing should still create Python Series
    s = Series.from_pylist(imgs, pyobj="force")
    assert s.datatype() == DataType.python()

    s = Series.from_pylist(imgs, pyobj="allow")
    assert s.datatype() == DataType.image()
    out = s.to_pylist()
    arrs[3] = np.expand_dims(arrs[3], axis=-1)
    np.testing.assert_equal(out, arrs)


@pytest.mark.parametrize(
    ["mode", "file_format"],
    [
        ("L", "png"),
        ("L", "tiff"),
        ("L", "jpeg"),
        ("LA", "png"),
        # Image crate doesn't support 2 samples per pixel.
        # ("LA", "tiff"),
        ("RGB", "png"),
        ("RGB", "tiff"),
        ("RGB", "bmp"),
        ("RGB", "gif"),
        ("RGB", "jpeg"),
        ("RGBA", "png"),
        ("RGBA", "tiff"),
        ("RGBA", "gif"),
        # Not supported by Daft or PIL.
        # "L16", "LA16", "RGB16", "RGBA16", "RGB32F", "RGBA32F"
    ],
)
@pytest.mark.parametrize("fixed_shape", [True, False])
def test_image_encode_pil(fixed_shape, mode, file_format):
    np_dtype = MODE_TO_NP_DTYPE[mode]
    num_channels = MODE_TO_NUM_CHANNELS[mode]
    if fixed_shape:
        height = 4
        width = 4
        dtype = DataType.image(mode, height, width)
        shape = (height, width)
        if num_channels > 1:
            shape += (num_channels,)
        arr = np.arange(np.prod(shape)).reshape(shape).astype(np_dtype)
        arrs = [arr, arr, None]
    else:
        dtype = DataType.image()
        shape1 = (2, 2)
        shape2 = (3, 3)
        if num_channels > 1:
            shape1 += (num_channels,)
            shape2 += (num_channels,)
        arr1 = np.arange(np.prod(shape1)).reshape(shape1).astype(np_dtype)
        arr2 = np.arange(np.prod(shape1), np.prod(shape1) + np.prod(shape2)).reshape(shape2).astype(np_dtype)
        arrs = [arr1, arr2, None]
    if mode in ("LA", "RGBA"):
        for arr in arrs:
            if arr is not None:
                arr[..., -1] = 255

    s = Series.from_pylist(arrs)
    t = s.cast(dtype)
    assert t.datatype() == dtype

    u = t.image.encode(file_format.upper())
    pil_imgs = [Image.open(io.BytesIO(bytes_)) if bytes_ is not None else None for bytes_ in u.to_pylist()]

    def pil_img_to_ndarray(img):
        if file_format == "gif":
            frames = [np.asarray(frame.copy().convert(mode), dtype=np.uint8) for frame in ImageSequence.Iterator(img)]
            if len(frames) == 1:
                return frames[0]
            else:
                return np.array(frames)
        else:
            return np.asarray(img)

    pil_decoded_imgs = [pil_img_to_ndarray(img) if img is not None else None for img in pil_imgs]
    if file_format == "jpeg":
        # Do lossy check; JPEG format is encoded at 0.75 quality.
        assert len(pil_decoded_imgs) == len(arrs)
        assert pil_decoded_imgs[-1] == arrs[-1]
        for pil_decoded_img, arr in zip(pil_decoded_imgs[:-1], arrs[:-1]):
            np.testing.assert_allclose(pil_decoded_img, arr, rtol=1, atol=4)
    else:
        np.testing.assert_equal(pil_decoded_imgs, arrs)


PIL_DECODE_MODES = [
    ("L", "png"),
    ("L", "tiff"),
    ("LA", "png"),
    # Image crate doesn't support 2 samples per pixel.
    # ("LA", "tiff"),
    ("RGB", "png"),
    ("RGB", "tiff"),
    ("RGB", "bmp"),
    ("RGBA", "png"),
    ("RGBA", "tiff"),
    # Not supported by Daft or PIL.
    # "L16", "LA16", "RGB16", "RGBA16", "RGB32F", "RGBA32F"
]


@pytest.mark.parametrize(
    ["mode", "file_format"],
    PIL_DECODE_MODES,
)
def test_image_decode_pil(mode, file_format):
    np_dtype = MODE_TO_NP_DTYPE[mode]
    num_channels = MODE_TO_NUM_CHANNELS[mode]
    shape = (4, 4)
    if num_channels > 1:
        shape += (num_channels,)
    arr = np.arange(np.prod(shape)).reshape(shape).astype(np_dtype)
    img = Image.fromarray(arr, mode=mode)
    img_bytes = io.BytesIO()
    img.save(img_bytes, file_format)
    img_bytes = img_bytes.getvalue()
    arrow_arr = pa.array([img_bytes, img_bytes, img_bytes], type=pa.binary())
    s = Series.from_arrow(arrow_arr)
    t = s.image.decode()
    # TODO(Clark): Infer type-leve mode if all images are the same mode.
    assert t.datatype() == DataType.image()
    out = t.cast(DataType.python()).to_pylist()
    expected_arrs = [arr, arr, arr]
    if num_channels == 1:
        expected_arrs = [np.expand_dims(arr, -1) for arr in expected_arrs]
    np.testing.assert_equal(out, expected_arrs)


@pytest.mark.parametrize("output_mode", ["L", "LA", "RGB", "RGBA"])
def test_image_decode_pil_multi_mode(output_mode):
    imgs = []
    for mode, file_format in PIL_DECODE_MODES:
        np_dtype = MODE_TO_NP_DTYPE[mode]
        num_channels = MODE_TO_NUM_CHANNELS[mode]
        shape = (4, 4)
        if num_channels > 1:
            shape += (num_channels,)
        arr = np.arange(np.prod(shape)).reshape(shape).astype(np_dtype)
        img = Image.fromarray(arr, mode=mode)
        img_bytes = io.BytesIO()
        img.save(img_bytes, file_format)
        img_bytes = img_bytes.getvalue()
        imgs.append(img_bytes)

    arrow_arr = pa.array(imgs, type=pa.binary())
    s = Series.from_arrow(arrow_arr)
    t = s.image.decode(mode=output_mode)
    assert t.datatype() == DataType.image(output_mode)
    out = t.cast(DataType.python()).to_pylist()
    for o in out:
        assert o.shape == (4, 4, MODE_TO_NUM_CHANNELS[output_mode])


@pytest.mark.parametrize(
    ["mode", "file_format"],
    [
        ("L", "png"),
        ("L", "tiff"),
        ("LA", "png"),
        # Image crate doesn't support 2 samples per pixel.
        # ("LA", "tiff"),
        ("RGB", "png"),
        ("RGB", "tiff"),
        ("RGB", "bmp"),
        ("RGBA", "png"),
        ("RGBA", "tiff"),
        # Not supported by Daft or PIL.
        # "L16", "LA16", "RGB16", "RGBA16", "RGB32F", "RGBA32F"
    ],
)
@pytest.mark.parametrize("fixed_shape", [True, False])
def test_image_encode_decode_pil_roundtrip(fixed_shape, mode, file_format):
    np_dtype = MODE_TO_NP_DTYPE[mode]
    num_channels = MODE_TO_NUM_CHANNELS[mode]
    if fixed_shape:
        height = 4
        width = 4
        shape = (height, width)
        if num_channels > 1:
            shape += (num_channels,)
        arr = np.arange(np.prod(shape)).reshape(shape).astype(np_dtype)
        arrs = [arr, arr, None]
    else:
        shape1 = (2, 2)
        shape2 = (3, 3)
        if num_channels > 1:
            shape1 += (num_channels,)
            shape2 += (num_channels,)
        arr1 = np.arange(np.prod(shape1)).reshape(shape1).astype(np_dtype)
        arr2 = np.arange(np.prod(shape1), np.prod(shape1) + np.prod(shape2)).reshape(shape2).astype(np_dtype)
        arrs = [arr1, arr2, None]
    if mode in ("LA", "RGBA"):
        for arr in arrs:
            if arr is not None:
                arr[..., -1] = 255
    imgs_bytes = []
    for arr in arrs:
        if arr is not None:
            img = Image.fromarray(arr, mode=mode)
            img_bytes = io.BytesIO()
            img.save(img_bytes, file_format)
            img_bytes = img_bytes.getvalue()
        else:
            img_bytes = None
        imgs_bytes.append(img_bytes)
    arrow_arr = pa.array(imgs_bytes, type=pa.binary())
    s = Series.from_arrow(arrow_arr)
    t = s.image.decode()
    # TODO(Clark): Infer type-leve mode if all images are the same mode.
    assert t.datatype() == DataType.image()

    u = t.image.encode(file_format.upper())
    pil_decoded_imgs = [
        np.asarray(Image.open(io.BytesIO(bytes_))) if bytes_ is not None else None for bytes_ in u.to_pylist()
    ]
    np.testing.assert_equal(pil_decoded_imgs, arrs)


@pytest.mark.parametrize(
    ["mode", "file_format"],
    [
        ("L", "png"),
        ("L", "tiff"),
        # OpenCV doesn't support 2-channel images.
        # ("LA", "png"),
        ("RGB", "png"),
        ("RGB", "tiff"),
        ("RGB", "bmp"),
        ("RGBA", "png"),
        ("RGBA", "tiff"),
        # Rust image crate doesn't support WebP encoding.
        # ("RGBA", "webp"),
        # TODO(Clark): Support uint16 images.
        # ("L16", "png"),
        # OpenCV doesn't support 2-channel images.
        # ("LA16", "png"),
        # TODO(Clark): Support uint16 images.
        # ("RGB16", "png"),
        # ("RGB16", "tiff"),
        # ("RGBA16", "png"),
        # ("RGBA16", "tiff"),
        # Image crate doesn't support LogLuv HDR encoding.
        # ("RGB32F", "tiff"),
        # ("RGBA32F", "tiff"),
    ],
)
def test_image_encode_opencv(mode, file_format):
    np_dtype = MODE_TO_NP_DTYPE[mode]
    num_channels = MODE_TO_NUM_CHANNELS[mode]
    shape = (4, 4, num_channels)
    arr = np.arange(np.prod(shape)).reshape(shape).astype(np_dtype)
    arrs = [arr, arr, arr]

    s = Series.from_pylist(arrs)
    t = s.cast(DataType.image(mode))
    assert t.datatype() == DataType.image(mode)
    # TODO(Clark): Support constructing an Image type with an unknown mode by known dtype.
    if np_dtype == np.uint8:
        # TODO(Clark): Infer type-leve mode if all images are the same mode.
        assert t.datatype() == DataType.image(mode)

    u = t.image.encode(file_format.upper())
    opencv_decoded_imgs = [
        cv2.imdecode(np.frombuffer(bytes_, dtype=np.uint8), cv2.IMREAD_UNCHANGED) for bytes_ in u.to_pylist()
    ]
    if num_channels == 1:
        opencv_decoded_imgs = [np.expand_dims(arr, -1) for arr in opencv_decoded_imgs]
    color_conv = MODE_TO_OPENCV_COLOR_CONVERSION_DECODE.get(mode)
    if color_conv is not None:
        opencv_decoded_imgs = [cv2.cvtColor(arr, color_conv) for arr in opencv_decoded_imgs]
    np.testing.assert_equal(opencv_decoded_imgs, arrs)


@pytest.mark.parametrize(
    ["mode", "file_format"],
    [
        ("L", "png"),
        ("L", "tiff"),
        # OpenCV doesn't support 2-channel images.
        # ("LA", "png"),
        ("RGB", "png"),
        ("RGB", "tiff"),
        ("RGB", "bmp"),
        ("RGBA", "png"),
        ("RGBA", "tiff"),
        ("RGBA", "webp"),
        # TODO(Clark): Support uint16 images.
        # ("L16", "png"),
        # OpenCV doesn't support 2-channel images.
        # ("LA16", "png"),
        # TODO(Clark): Support uint16 images.
        # ("RGB16", "png"),
        # ("RGB16", "tiff"),
        # ("RGBA16", "png"),
        # ("RGBA16", "tiff"),
        # Image crate doesn't support LogLuv HDR encoding.
        # ("RGB32F", "tiff"),
        # ("RGBA32F", "tiff"),
    ],
)
def test_image_decode_opencv(mode, file_format):
    np_dtype = MODE_TO_NP_DTYPE[mode]
    num_channels = MODE_TO_NUM_CHANNELS[mode]
    shape = (4, 4, num_channels)
    arr = np.arange(np.prod(shape)).reshape(shape).astype(np_dtype)
    cv2_arr = arr
    color_conv = MODE_TO_OPENCV_COLOR_CONVERSION.get(mode)
    if color_conv is not None:
        cv2_arr = cv2.cvtColor(arr, color_conv)
    encoded_arr = cv2.imencode(f".{file_format}", cv2_arr)[1]
    img_bytes = encoded_arr.tobytes()
    arrow_arr = pa.array([img_bytes, img_bytes, img_bytes], type=pa.binary())
    s = Series.from_arrow(arrow_arr)
    t = s.image.decode()
    # TODO(Clark): Support constructing an Image type with an unknown mode by known dtype.
    if np_dtype == np.uint8:
        # TODO(Clark): Infer type-leve mode if all images are the same mode.
        assert t.datatype() == DataType.image()
    out = t.cast(DataType.python()).to_pylist()
    expected_arrs = [arr, arr, arr]
    np.testing.assert_equal(out, expected_arrs)


# check support for decoding even from unsupported types
@pytest.mark.parametrize("output_mode", ["L", "LA", "RGB", "RGBA"])
def test_image_decode_opencv_multi_modes(output_mode):
    INPUT_MODES = [
        ("L", "png"),
        ("L", "tiff"),
        # OpenCV doesn't support 2-channel images.
        # ("LA", "png"),
        ("RGB", "png"),
        ("RGB", "tiff"),
        ("RGB", "bmp"),
        ("RGBA", "png"),
        ("RGBA", "tiff"),
        ("RGBA", "webp"),
        ("L16", "png"),
        # OpenCV doesn't support 2-channel images.
        # ("LA16", "png"),
        ("RGB16", "png"),
        ("RGB16", "tiff"),
        ("RGBA16", "png"),
        ("RGBA16", "tiff"),
        # Image crate doesn't support LogLuv HDR encoding.
        # ("RGB32F", "tiff"),
        # ("RGBA32F", "tiff"),
    ]
    imgs = []
    for mode, file_format in INPUT_MODES:
        np_dtype = MODE_TO_NP_DTYPE[mode]
        num_channels = MODE_TO_NUM_CHANNELS[mode]
        shape = (4, 4, num_channels)
        arr = np.arange(np.prod(shape)).reshape(shape).astype(np_dtype)
        cv2_arr = arr
        color_conv = MODE_TO_OPENCV_COLOR_CONVERSION.get(mode)
        if color_conv is not None:
            cv2_arr = cv2.cvtColor(arr, color_conv)
        encoded_arr = cv2.imencode(f".{file_format}", cv2_arr)[1]
        img_bytes = encoded_arr.tobytes()
        imgs.append(img_bytes)
    arrow_arr = pa.array(imgs, type=pa.binary())
    s = Series.from_arrow(arrow_arr)
    t = s.image.decode(mode=output_mode)
    assert t.datatype() == DataType.image(output_mode)
    out = t.to_pylist()
    for o in out:
        assert o.shape == (4, 4, MODE_TO_NUM_CHANNELS[output_mode])


@pytest.mark.parametrize(
    ["mode", "file_format"],
    [
        ("L", "png"),
        ("L", "tiff"),
        # OpenCV doesn't support 2-channel images.
        # ("LA", "png"),
        ("RGB", "png"),
        ("RGB", "tiff"),
        ("RGB", "bmp"),
        ("RGBA", "png"),
        ("RGBA", "tiff"),
        # Rust image crate doesn't support WebP encoding.
        # ("RGBA", "webp"),
        # TODO(Clark): Support uint16 images.
        # ("L16", "png"),
        # OpenCV doesn't support 2-channel images.
        # ("LA16", "png"),
        # TODO(Clark): Support uint16 images.
        # ("RGB16", "png"),
        # ("RGB16", "tiff"),
        # ("RGBA16", "png"),
        # ("RGBA16", "tiff"),
        # Image crate doesn't support LogLuv HDR encoding.
        # ("RGB32F", "tiff"),
        # ("RGBA32F", "tiff"),
    ],
)
def test_image_encode_decode_opencv_roundtrip(mode, file_format):
    np_dtype = MODE_TO_NP_DTYPE[mode]
    num_channels = MODE_TO_NUM_CHANNELS[mode]
    shape = (4, 4, num_channels)
    arr = np.arange(np.prod(shape)).reshape(shape).astype(np_dtype)
    cv2_arr = arr
    color_conv = MODE_TO_OPENCV_COLOR_CONVERSION.get(mode)
    if color_conv is not None:
        cv2_arr = cv2.cvtColor(arr, color_conv)
    encoded_arr = cv2.imencode(f".{file_format}", cv2_arr)[1]
    img_bytes = encoded_arr.tobytes()
    arrow_arr = pa.array([img_bytes, img_bytes, img_bytes], type=pa.binary())

    s = Series.from_arrow(arrow_arr)
    t = s.image.decode()
    # TODO(Clark): Support constructing an Image type with an unknown mode by known dtype.
    if np_dtype == np.uint8:
        # TODO(Clark): Infer type-leve mode if all images are the same mode.
        assert t.datatype() == DataType.image()

    u = t.image.encode(file_format.upper())
    opencv_decoded_imgs = [
        cv2.imdecode(np.frombuffer(bytes_, dtype=np.uint8), cv2.IMREAD_UNCHANGED) for bytes_ in u.to_pylist()
    ]
    if num_channels == 1:
        opencv_decoded_imgs = [np.expand_dims(arr, -1) for arr in opencv_decoded_imgs]
    color_conv = MODE_TO_OPENCV_COLOR_CONVERSION_DECODE.get(mode)
    if color_conv is not None:
        opencv_decoded_imgs = [cv2.cvtColor(arr, color_conv) for arr in opencv_decoded_imgs]
    np.testing.assert_equal(opencv_decoded_imgs, [arr, arr, arr])


@pytest.mark.parametrize("mode", ["L", "LA", "RGB", "RGBA"])
@pytest.mark.parametrize("fixed_shape", [True, False])
def test_image_resize_same_mode(fixed_shape, mode):
    np_dtype = MODE_TO_NP_DTYPE[mode]
    num_channels = MODE_TO_NUM_CHANNELS[mode]
    if fixed_shape:
        height = 3
        width = 4
        dtype = DataType.image(mode, height, width)
        shape = (height, width, num_channels)
        arr = np.arange(np.prod(shape)).reshape(shape).astype(np_dtype)
        arrs = [arr, arr, None]
    else:
        dtype = DataType.image(mode)
        shape1 = (2, 3, num_channels)
        shape2 = (3, 4, num_channels)
        arr1 = np.arange(np.prod(shape1)).reshape(shape1).astype(np_dtype)
        arr2 = np.arange(np.prod(shape1), np.prod(shape1) + np.prod(shape2)).reshape(shape2).astype(np_dtype)
        arrs = [arr1, arr2, None]

    if mode in ("LA", "RGBA"):
        for arr in arrs:
            if arr is not None:
                arr[..., -1] = 255

    s = Series.from_pylist(arrs, pyobj="force")
    t = s.cast(dtype)
    assert t.datatype() == dtype

    resized = t.image.resize(5, 5)
    resized_dtype = DataType.image(mode, 5, 5)
    assert resized.datatype() == resized_dtype
    out = resized.cast(DataType.python()).to_pylist()

    def resize(arr):
        # Use opencv as a resizing baseline.
        return cv2.resize(arr, dsize=(5, 5), interpolation=cv2.INTER_LINEAR_EXACT)

    expected = [resize(arr) if arr is not None else None for arr in arrs]
    if num_channels == 1:
        expected = [np.expand_dims(arr, -1) for arr in expected]

    np.testing.assert_equal(out, expected)


def test_image_resize_mixed_modes():
    rgba = np.ones((2, 2, 4), dtype=np.uint8)
    rgba[..., 1] = 2
    rgba[..., 2] = 3
    rgba[..., 3] = 4

    data = [
        rgba[..., :3],  # rgb
        rgba,  # RGBA
        np.arange(12, dtype=np.uint8).reshape((1, 4, 3)),  # RGB
        np.arange(12, dtype=np.uint8).reshape((3, 4)) * 10,  # L
        np.ones(24, dtype=np.uint8).reshape((3, 4, 2)) * 10,  # LA
        None,
    ]

    s = Series.from_pylist(data, pyobj="force")

    target_dtype = DataType.image()

    t = s.cast(target_dtype)

    assert t.datatype() == target_dtype

    resized = t.image.resize(5, 5)

    out = resized.cast(DataType.python()).to_pylist()

    def resize(arr):
        # Use opencv as a resizing baseline.
        arr = cv2.resize(arr, dsize=(5, 5), interpolation=cv2.INTER_LINEAR_EXACT)
        if arr.ndim == 2:
            arr = np.expand_dims(arr, -1)
        return arr

    np.testing.assert_equal(out, [resize(arr) if arr is not None else None for arr in data])


def test_fixed_shape_image_roundtrip():
    height = 2
    width = 2
    shape = (height, width, 3)
    data = [np.arange(12, dtype=np.uint8).reshape(shape), np.arange(12, 24, dtype=np.uint8).reshape(shape), None]
    s = Series.from_pylist(data, pyobj="force")

    target_dtype = DataType.image("RGB", height, width)

    t = s.cast(target_dtype)

    assert t.datatype() == target_dtype

    # Test pylist roundtrip.
    back_dtype = DataType.python()
    back = t.cast(back_dtype)

    assert back.datatype() == back_dtype

    out = back.to_pylist()
    np.testing.assert_equal(out, data)

    # Test Arrow roundtrip.
    arrow_arr = t.to_arrow()

    assert isinstance(arrow_arr.type, DaftExtension)
    from_arrow = Series.from_arrow(t.to_arrow())

    assert from_arrow.datatype() == t.datatype()
    np.testing.assert_equal(from_arrow.to_pylist(), t.to_pylist())

    t_copy = copy.deepcopy(t)
    assert t_copy.datatype() == t.datatype()
    np.testing.assert_equal(t_copy.to_pylist(), t.to_pylist())


def test_bad_cast_image():
    data = [
        np.arange(12, dtype=np.uint8).reshape((2, 2, 3)),
        np.arange(12, 39, dtype=np.uint64).reshape((3, 3, 3)),
        None,
    ]
    s = Series.from_pylist(data, pyobj="force")

    target_dtype = DataType.image("RGB")
    with pytest.raises(ValueError, match="Expected Numpy array to be of type: UInt8"):
        s.cast(target_dtype)


# Add enforcement for fixed sized list
@pytest.mark.skip()
def test_bad_cast_fixed_shape_image():
    height = 2
    width = 2
    shape = (height, width, 3)
    data = [np.arange(12, dtype=np.uint8).reshape(shape), np.arange(12, 24, dtype=np.uint64).reshape(shape), None]
    s = Series.from_pylist(data, pyobj="force")

    target_dtype = DataType.image("RGB", height, width)

    with pytest.raises(ValueError, match="Expected Numpy array to be of type: UInt8"):
        s.cast(target_dtype)


def test_on_error_image_decode():
    s = Series.from_pylist([b"not an image"])

    with pytest.raises(Exception, match="Invalid on_error value: NOT IMPLEMENTED"):
        s.image.decode(on_error="NOT IMPLEMENTED")

    with pytest.raises(ValueError, match="Decoding image from bytes failed"):
        s.image.decode(on_error="raise")

    s.image.decode(on_error="null").to_pylist() == [None]


mode_cast_modes = ["L", "LA", "RGB", "RGBA"]


@pytest.mark.parametrize(
    ["input_mode", "output_mode"],
    [(a, b) for a in mode_cast_modes for b in mode_cast_modes],
)
def test_image_to_mode(input_mode, output_mode):
    channels = MODE_TO_NUM_CHANNELS[input_mode]
    data = [
        np.arange(4 * channels, dtype=np.uint8).reshape((2, 2, channels)),
        np.arange(4 * channels, 13 * channels, dtype=np.uint8).reshape((3, 3, channels)),
        None,
    ]
    s = Series.from_pylist(data, pyobj="force")

    s = s.cast(DataType.image(input_mode)).image.to_mode(output_mode)
    assert s.datatype() == DataType.image(output_mode)
    assert s.to_pylist()[0].shape[2] == MODE_TO_NUM_CHANNELS[output_mode]


@pytest.mark.parametrize(
    ["input_mode", "output_mode"],
    [(a, b) for a in mode_cast_modes for b in mode_cast_modes],
)
def test_image_to_mode_fixed_size(input_mode, output_mode):
    channels = MODE_TO_NUM_CHANNELS[input_mode]
    data = [
        np.arange(4 * channels, dtype=np.uint8).reshape((2, 2, channels)),
        np.arange(4 * channels, 8 * channels, dtype=np.uint8).reshape((2, 2, channels)),
        None,
    ]
    s = Series.from_pylist(data, pyobj="force")

    s = s.cast(DataType.image(input_mode, 2, 2)).image.to_mode(output_mode)
    assert s.datatype() == DataType.image(output_mode, 2, 2)
    assert s.to_pylist()[0].shape[2] == MODE_TO_NUM_CHANNELS[output_mode]



from __future__ import annotations

from typing import TYPE_CHECKING

from daft.daft import PyPushdowns
from daft.expressions import col, lit
from daft.io.pushdowns import Pushdowns

if TYPE_CHECKING:
    from daft.expressions import Expression


###
# Pushdowns Translations Tests
###


def assert_eq(e1: Expression, e2: Expression):
    assert e1._expr._eq(e2._expr)


def test_projection_pushdowns():
    py_pushdowns = PyPushdowns(columns=["a", "b", "c"])
    pushdowns = Pushdowns._from_pypushdowns(py_pushdowns)

    assert pushdowns.columns
    assert len(pushdowns.columns) == 3
    assert pushdowns.columns[0] == "a"
    assert pushdowns.columns[1] == "b"
    assert pushdowns.columns[2] == "c"

    assert pushdowns.limit is None
    assert pushdowns.filters is None


def test_simple_filters_pushdown():
    filters = col("a") == lit(1)  # (= a 1)
    py_pushdowns = PyPushdowns(filters=filters._expr)
    pushdowns = Pushdowns._from_pypushdowns(py_pushdowns)

    assert_eq(pushdowns.filters, filters)  # <- roundtrip assert

    assert pushdowns.columns is None
    assert pushdowns.limit is None


def test_complex_filters_pushdown():
    filters = col("a") == (col("b") + col("c"))  # (= a (+ b c))
    py_pushdowns = PyPushdowns(filters=filters._expr)
    pushdowns = Pushdowns._from_pypushdowns(py_pushdowns)

    assert_eq(pushdowns.filters, filters)  # <- roundtrip assert

    assert pushdowns.columns is None
    assert pushdowns.limit is None


def test_limit_pushdown():
    py_pushdowns = PyPushdowns(limit=1738)
    pushdowns = Pushdowns._from_pypushdowns(py_pushdowns)

    assert pushdowns.columns is None
    assert pushdowns.filters is None
    assert pushdowns.limit == 1738


def test_simple_partition_pushdown():
    partition_filters = col("a") == lit(1)  # (= a 1)
    py_pushdowns = PyPushdowns(partition_filters=partition_filters._expr)
    pushdowns = Pushdowns._from_pypushdowns(py_pushdowns)

    # roundtrip assert
    assert_eq(pushdowns.partition_filters, partition_filters)

    assert pushdowns.filters is None
    assert pushdowns.columns is None
    assert pushdowns.limit is None


def test_composite_predicate_pushdown():
    filters = col("a") == lit(1)  # (= a 1)
    partition_filters = col("b") > lit(2)  # (> b 2)
    py_pushdowns = PyPushdowns(filters=filters._expr, partition_filters=partition_filters._expr)
    pushdowns = Pushdowns._from_pypushdowns(py_pushdowns)

    # roundtrip assert
    assert_eq(pushdowns.filters, filters)
    assert_eq(pushdowns.partition_filters, partition_filters)

    assert pushdowns.columns is None
    assert pushdowns.limit is None



from __future__ import annotations

import pytest

import daft
from daft import Session


def _df(name):
    """Create some table whose only column the case-preserved name so we know exactly which was found."""
    return daft.from_pydict({name: []})


def assert_ok(sess: Session, ident: str, find: str):
    """Assert that 'ident' is resolved to 'find'."""
    assert sess.sql(f"SELECT * FROM {ident.strip()}").to_pydict() == {find: []}


def assert_err(sess: Session, ident: str, match: str):
    """Assert that 'ident' fails to resolve with the given message match."""
    with pytest.raises(Exception, match=match):
        sess.sql(f"SELECT * FROM {ident}")


def test_case_sensitive_identifiers():
    sess = Session()

    # unambiguous uppercase name
    sess.create_temp_table("T", _df("T"))

    # ok since regular and delimited are both case-sensitive.
    assert_ok(sess, "  T  ", find="T")
    assert_ok(sess, ' "T" ', find="T")

    # fails since case-sensitive cannot find 'T'
    assert_err(sess, "  t  ", match="Table not found")
    assert_err(sess, ' "t" ', match="Table not found")
    # unambiguous lowercase name
    sess.create_temp_table("s", _df("s"))

    # ok since regular and delimited are both case-sensitive.
    assert_ok(sess, "  s  ", find="s")
    assert_ok(sess, ' "s" ', find="s")

    # fails since case-sensitive cannot find 's'
    assert_err(sess, "  S  ", match="Table not found")
    assert_err(sess, ' "S" ', match="Table not found")

    # possibly ambiguous names depending on mode
    sess.create_temp_table("abc", _df("abc"))
    sess.create_temp_table("ABC", _df("ABC"))
    sess.create_temp_table("aBc", _df("aBc"))

    # delimited identifiers are case-sensitive
    assert_ok(sess, ' "abc" ', find="abc")
    assert_ok(sess, ' "ABC" ', find="ABC")
    assert_ok(sess, ' "aBc" ', find="aBc")

    # regular identifiers are also case-sensitive
    assert_ok(sess, " abc ", find="abc")
    assert_ok(sess, " ABC ", find="ABC")
    assert_ok(sess, " aBc ", find="aBc")
